from tensorflow.python.keras.layers import (
    ZeroPadding2D,
    Conv2D,
    Activation,
    LeakyReLU,
    ZeroPadding3D,
    Conv3D,
    Cropping2D,
    Cropping3D,
    AveragePooling2D,
    MaxPooling2D,
    AveragePooling3D,
    MaxPooling3D,
    BatchNormalization,
)
from tensorflow.python.keras.regularizers import l1

from aydin.nn.layers.instance_norm import InstanceNormalization
from aydin.nn.layers.util import Swish


def conv2d_torch(
    x,
    out_channels,
    kernel_size,
    norm=None,
    act='ReLU',
    lyrname='cv2dtrch',
    padding=0,
    dilation_rate=1,
    bias=True,
    leaky_alpha=0.01,
):
    """
    Parameters
    ----------
    x
    out_channels
    kernel_size
    norm
    act
    lyrname
    padding
    dilation_rate
    bias
    leaky_alpha

    Returns
    -------
    conv2d_torch layer

    """
    x1 = ZeroPadding2D(padding, name=lyrname + '_pd0')(x)

    x1 = Conv2D(
        out_channels,
        kernel_size,
        dilation_rate=dilation_rate,
        name=lyrname + '_cv',
        use_bias=bias,
    )(x1)

    if norm == 'instance':
        x1 = InstanceNormalization(name=lyrname + '_in')(x1)
    elif norm == 'batch':
        x1 = BatchNormalization(name=lyrname + '_bn')(x1)
    if act == 'ReLU':
        return Activation('relu', name=lyrname + '_relu')(x1)
    elif act == 'swish':
        return Swish(name=lyrname + '_swsh')(x1)
    elif act in ['lrel', 'leaky', 'LeakyReLU']:
        return LeakyReLU(alpha=leaky_alpha, name=lyrname + '_lrel')(x1)
    else:
        return x1


def conv3d_torch(
    x,
    out_channels,
    kernel_size,
    norm=None,
    act='ReLU',
    lyrname='cv3dtrch',
    padding=0,
    dilation_rate=1,
    bias=True,
    leaky_alpha=0.01,
):
    """

    Parameters
    ----------
    x
    out_channels
    kernel_size
    norm
    act
    lyrname
    padding
    dilation_rate
    bias
    leaky_alpha

    Returns
    -------
    conv3d_torch layer

    """
    x1 = ZeroPadding3D(padding, name=lyrname + '_pd0')(x)

    x1 = Conv3D(
        out_channels,
        kernel_size,
        dilation_rate=dilation_rate,
        name=lyrname + '_cv',
        use_bias=bias,
    )(x1)

    if norm == 'instance':
        x1 = InstanceNormalization(name=lyrname + '_in')(x1)
    elif norm == 'batch':
        x1 = BatchNormalization(name=lyrname + '_bn')(x1)
    if act == 'ReLU':
        return Activation('relu', name=lyrname + '_relu')(x1)
    elif act == 'swish':
        return Swish(name=lyrname + '_swsh')(x1)
    elif act in ['lrel', 'leaky', 'LeakyReLU']:
        return LeakyReLU(alpha=leaky_alpha, name=lyrname + '_lrel')(x1)
    else:
        return x1


def conv2d_bn(
    xx,
    unit,
    kernel_size=3,
    shiftconv=False,
    norm=None,
    act='ReLU',
    weight_decay=0,
    lyrname=None,
):
    """

    Parameters
    ----------
    xx
    unit
    kernel_size
    shiftconv
    norm
    act
    weight_decay
    lyrname

    Returns
    -------
    conv2d_bn layer

    """
    if shiftconv:
        x1 = ZeroPadding2D(((0, 0), (1, 0)), name=lyrname + '_0pd')(xx)
        x1 = Conv2D(unit, (3, 3), padding='same', name=lyrname + '_cv2')(x1)
        x1 = Cropping2D(((0, 0), (0, 1)), name=lyrname + '_crp')(x1)
    else:
        x1 = Conv2D(
            unit,
            kernel_size,
            padding='same',
            kernel_regularizer=l1(weight_decay),
            bias_regularizer=l1(weight_decay),
            name=lyrname + '_cv2',
        )(xx)

    if norm == 'instance':
        x1 = InstanceNormalization(name=lyrname + '_in')(x1)
    elif norm == 'batch':
        x1 = BatchNormalization(name=lyrname + '_bn')(x1)
    if act == 'ReLU':
        return Activation('relu', name=lyrname + '_relu')(x1)
    elif act == 'swish':
        return Swish(name=lyrname + '_swsh')(x1)
    elif act == 'lrel':
        return LeakyReLU(alpha=0.1, name=lyrname + '_lrel')(x1)
    else:
        return x1


def conv3d_bn(
    xx,
    unit,
    kernel_size=3,
    shiftconv=False,
    norm=None,
    act='ReLU',
    weight_decay=0,
    lyrname=None,
):
    """

    Parameters
    ----------
    xx
    unit
    kernel_size
    shiftconv
    norm
    act
    weight_decay
    lyrname

    Returns
    -------
    conv3d_bn layer

    """
    if shiftconv:
        x1 = ZeroPadding3D(((0, 0), (0, 0), (1, 0)), name=lyrname + '_0pd')(xx)
        x1 = Conv3D(unit, (3, 3, 3), padding='same', name=lyrname + '_cv')(x1)
        x1 = Cropping3D(((0, 0), (0, 0), (0, 1)), name=lyrname + '_crp')(x1)
    else:
        x1 = Conv3D(
            unit,
            kernel_size,
            padding='same',
            kernel_regularizer=l1(weight_decay),
            bias_regularizer=l1(weight_decay),
            name=lyrname + '_cv3',
        )(xx)
    if norm == 'instance':
        x1 = InstanceNormalization(name=lyrname + '_in')(x1)
    elif norm == 'batch':
        x1 = BatchNormalization(name=lyrname + '_bn')(x1)
    if act == 'ReLU':
        return Activation('relu', name=lyrname + '_relu')(x1)
    elif act == 'swish':
        return Swish(name=lyrname + '_swsh')(x1)
    elif act == 'lrel':
        return LeakyReLU(alpha=0.1, name=lyrname + '_lrel')(x1)
    else:
        return x1


def pooling_down2D(xx, shiftconv, mode='max', lyrname=None):
    """

    Parameters
    ----------
    xx
    shiftconv
    mode
    lyrname

    Returns
    -------
    pooling_down2D layer

    """
    if shiftconv:
        x1 = ZeroPadding2D(((0, 0), (1, 0)), name=lyrname + '_0pd')(xx)
        x1 = Cropping2D(((0, 0), (0, 1)), name=lyrname + '_crp')(x1)
    else:
        x1 = xx
    if mode == 'ave':
        x1 = AveragePooling2D((2, 2), name=lyrname + '_apl')(x1)
    elif mode == 'max':
        x1 = MaxPooling2D((2, 2), name=lyrname + '_mpl')(x1)
    else:
        raise ValueError('pooling mode only accepts "max" or "ave".')
    return x1


def pooling_down3D(xx, shiftconv, pool_size=(2, 2, 2), mode='max', lyrname=None):
    """

    Parameters
    ----------
    xx
    shiftconv
    pool_size
    mode
    lyrname

    Returns
    -------
    pooling_down3D layer

    """
    if shiftconv:
        x1 = ZeroPadding3D(((0, 0), (0, 0), (1, 0)), name=lyrname + '_0pd')(xx)
        x1 = Cropping3D(((0, 0), (0, 0), (0, 1)), name=lyrname + '_crp')(x1)
    else:
        x1 = xx
    if mode == 'ave':
        x1 = AveragePooling3D(pool_size, name=lyrname + '_apl')(x1)
    elif mode == 'max':
        x1 = MaxPooling3D(pool_size, name=lyrname + '_mpl')(x1)
    else:
        raise ValueError('pooling mode only accepts "max" or "ave".')
    return x1



import numpy
from skimage.exposure import rescale_intensity

from aydin.features.extensible_features import ExtensibleFeatureGenerator
from aydin.features.groups.median import MedianFeatures
from aydin.features.groups.translations import TranslationFeatures
from aydin.io.datasets import camera


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def test_extensible_features_basics():
    image = n(camera().astype(numpy.float32))

    # feature generator requires images in 'standard' form: BCTZYX
    image = image[numpy.newaxis, numpy.newaxis, ...]

    feature_gen = ExtensibleFeatureGenerator()

    vectors = [(0, -1), (0, 1), (-1, 0), (1, 0), (-1, -1), (-1, 1), (1, -1), (1, 1)]
    translations = TranslationFeatures(translations=vectors)
    feature_gen.add_feature_group(translations)

    radii = [1, 2, 4, 8]
    medians = MedianFeatures(radii=radii)
    feature_gen.add_feature_group(medians)

    features = feature_gen.compute(image, exclude_center_feature=True)

    assert feature_gen.get_num_features(image.ndim - 2) == 8 + 4
    assert feature_gen.get_receptive_field_radius() == 8

    assert features is not None
    assert features.shape == (1, 512, 512, 12)



def closest_product(values, N, tol_min=-0.1, tol_max=0.1):
    """
    Given a list of integers u=[a_0, ..., a_m]  find the set of indices
    {i_0, ..., i_n} for that list such that the product u[i_0]* ... *u[i_n]
    is the closest to a given integer N.

    This method is pyramid heuristic approximation that returns the closest
    solution within N*[tol_min,tol_max]

    Parameters
    ----------
    values : array_like
    N : int
    tol_min : float
    tol_max : float

    Returns
    -------
    List of ints or None

    """
    length = len(values)

    # it is just there...
    if N in values:
        return [values.index(N)]

    # let's look at pairs of values:
    for a in range(0, length):
        for b in range(a + 1, length):
            value = abs(values[a] * values[b] - N) / N
            if tol_min <= value and value < tol_max:
                return [a, b]

    # let's look at triplets of values:
    for a in range(0, length):
        for b in range(a + 1, length):
            for c in range(b + 1, length):
                value = abs(values[a] * values[b] * values[c] - N) / N
                if tol_min <= value and value < tol_max:
                    return [a, b, c]

    # let's look at quadruplets of values:
    for a in range(0, length):
        for b in range(a + 1, length):
            for c in range(b + 1, length):
                for d in range(c + 1, length):
                    value = abs(values[a] * values[b] * values[c] * values[d] - N) / N
                    if tol_min <= value and value < tol_max:
                        return [a, b, c, d]

    # let's look at quintuplets of values:
    for a in range(0, length):
        for b in range(a + 1, length):
            for c in range(b + 1, length):
                for d in range(c + 1, length):
                    for e in range(d + 1, length):
                        value = (
                            abs(
                                values[a]
                                * values[b]
                                * values[c]
                                * values[d]
                                * values[e]
                                - N
                            )
                            / N
                        )
                        if tol_min <= value and value < tol_max:
                            return [a, b, c, d, e]

    # There is no chance we need more...

    # Given the constraints, it might not be possible to find a soliution anyway.
    # in that case we give up and return None:
    return None



# flake8: noqa
import time

import numpy
from skimage.exposure import rescale_intensity
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.analysis.image_metrics import mutual_information, spectral_mutual_information
from aydin.io.datasets import normalise, examples_single, add_blur_3d
from aydin.it.deconvolution.lr_deconv_cupy import ImageTranslatorLRDeconvCupy
from aydin.util.log.log import Log


def demo(image):
    Log.enable_output = True

    image = normalise(image.astype(numpy.float32))
    # image = image[100:165, 133:217, 150:231]
    blurred_image, psf_kernel = add_blur_3d(image, xy_size=17, z_size=31)

    it = ImageTranslatorLRDeconvCupy(
        psf_kernel=psf_kernel, max_num_iterations=100, clip=False
    )

    start = time.time()
    it.train(blurred_image)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    start = time.time()
    lr_deconvolved_image = it.translate(blurred_image)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    print("Below in order: PSNR, norm spectral mutual info, norm mutual info, SSIM: ")

    def printscore(header, val1, val2, val3, val4):
        print(f"{header}: \t {val1:.4f} \t {val2:.4f} \t {val3:.4f} \t {val4:.4f}")

    printscore(
        "n&b image",
        psnr(image, blurred_image),
        spectral_mutual_information(image, blurred_image),
        mutual_information(image, blurred_image),
        ssim(image, blurred_image),
    )

    printscore(
        "lr      ",
        psnr(image, lr_deconvolved_image),
        spectral_mutual_information(image, lr_deconvolved_image),
        mutual_information(image, lr_deconvolved_image),
        ssim(image, lr_deconvolved_image),
    )

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(psf_kernel, name='psf_kernel')
        viewer.add_image(image, name='image')
        viewer.add_image(blurred_image, name='blurred')
        viewer.add_image(lr_deconvolved_image, name='lr_deconvolved_image')


image = examples_single.janelia_flybrain.get_array()
image = image[:, 1, :, :].astype(numpy.float32)
# image = examples_single.keller_dmel.get_array()
# image = image.astype(numpy.float32)
image = rescale_intensity(image, in_range='image', out_range=(0, 1))

demo(image)



from pathlib import Path
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QLabel,
    QFileDialog,
    QTreeWidgetItem,
    QTreeWidget,
    QHBoxLayout,
    QPushButton,
    QCheckBox,
)

from aydin.gui._qt.custom_widgets.horizontal_line_break_widget import (
    QHorizontalLineBreakWidget,
)
from aydin.gui._qt.qtreewidget_utils import iter_tree_widget


class FilesTab(QWidget):
    """
    The first step is to load image files into Aydin.

    Drag and drop files or select them with the 'Add Files' button. You can see the files added, and remove files if
    needed. If a single file consists of multiple channels, you can split the image into separate images,
    one per channel, by ticking the corresponding box. If all the images added have the same shape and data type you
    have the option to 'hyperstack' all listed images together along a new dimension. Hyperstacking is selected by
    default if the image shapes and data types are compatible.
    """

    def __init__(self, parent):
        super(QWidget, self).__init__(parent)
        self.parent = parent

        self._hyperstack_last_fail_state_metadatas = None

        self.tab_layout = QVBoxLayout()

        # Explanation text
        self.explanation_text = QLabel(self.__doc__, self)
        self.tab_layout.addWidget(self.explanation_text)

        # Horizontal Line Break
        self.tab_layout.addWidget(QHorizontalLineBreakWidget(self))

        self.label_and_removebutton_layout = QHBoxLayout()
        label = QLabel("Click on column header to check/uncheck all rows below.", self)

        self.remove_button = QPushButton("Remove")
        self.remove_button.setStyleSheet("background-color: purple;")
        self.remove_button.clicked.connect(self.remove_selected_items_from_tree)
        self.remove_all_button = QPushButton("Remove All")
        self.remove_all_button.setStyleSheet("background-color: purple;")
        self.remove_button_layout = QHBoxLayout()
        self.remove_button_layout.setAlignment(Qt.AlignRight)
        self.remove_button_layout.addWidget(self.remove_button)
        self.remove_button_layout.addWidget(self.remove_all_button)

        self.label_and_removebutton_layout.addWidget(label)
        self.label_and_removebutton_layout.addLayout(self.remove_button_layout)
        self.tab_layout.addLayout(self.label_and_removebutton_layout)

        self.file_list_tree_widget = QTreeWidget()
        self.file_list_tree_widget.setHeaderLabels(
            ['File Name', 'Split Channels', 'axes', 'shape', 'dtype', 'Path']
        )
        self.file_list_tree_widget.header().sectionClicked.connect(
            self.onSectionClicked
        )
        self.file_list_tree_widget.setColumnWidth(0, 400)
        self.file_list_tree_widget.setColumnWidth(3, 400)
        self.file_list_tree_widget.header().setSectionsClickable(True)
        self.file_list_tree_widget.dropEvent = self.dropEvent
        self.file_list_tree_widget.dragMoveEvent = self.dragMoveEvent
        self.file_list_tree_widget.dragEnterEvent = self.dragEnterEvent
        self.file_list_tree_widget.itemChanged.connect(self.onTreeItemChanged)
        self.remove_all_button.clicked.connect(self.remove_all_items_from_tree)

        self.tab_layout.addWidget(self.file_list_tree_widget)

        # Checkbox of hyperstacking
        self.hyperstack_checkbox = QCheckBox(
            "Merge files into a single image(Hyperstack)", self
        )
        self.hyperstack_checkbox.toggled.connect(self.on_hyperstack_checkbox_toggled)
        self.tab_layout.addWidget(self.hyperstack_checkbox)

        self.tab_layout.setAlignment(Qt.AlignTop)
        self.setLayout(self.tab_layout)

    @Slot(int)
    def onSectionClicked(self, column):
        if column == 0:
            return
        for idx, item in enumerate(
            iter_tree_widget(self.file_list_tree_widget.invisibleRootItem())
        ):
            if idx == 0:
                continue
            elif idx == 1:
                state_to_be_set = (
                    Qt.Unchecked
                    if item.checkState(column) == Qt.Checked
                    else Qt.Checked
                )
                # self.check_for_split_changes()

            item.setCheckState(column, state_to_be_set)

    def onTreeItemChanged(self, item, column):
        if column == 1:
            if (
                self.parent.data_model.set_split_channels(
                    item.text(0), item.text(5), item.checkState(column)
                )
                == -1
            ):
                item.setCheckState(column, False)

    def remove_selected_items_from_tree(self):
        root = self.file_list_tree_widget.invisibleRootItem()
        for item in self.file_list_tree_widget.selectedItems():
            self.parent.data_model.remove_filepaths([item.text(5)])

            (item.parent() or root).removeChild(item)

    def remove_all_items_from_tree(self):
        self.file_list_tree_widget.clear()
        self.parent.data_model.clear_filepaths()

    def openFileNamesDialog(self):
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        files, _ = QFileDialog.getOpenFileNames(
            self, "Open File(s)", "", "All Files (*)", options=options
        )

        if files:
            self.parent.data_model.add_filepaths(files)
            self.parent.tabwidget.setCurrentIndex(1)

    def on_hyperstack_checkbox_toggled(self):
        response = self.parent.data_model.set_hyperstack(
            self.hyperstack_checkbox.isChecked()
        )
        if response == -1:
            self.hyperstack_checkbox.setChecked(False)

    def on_data_model_update(self):
        filepaths = self.parent.data_model.filepaths

        self.file_list_tree_widget.clear()

        if len(filepaths) == 0:
            return

        for path, (array, metadata) in filepaths.items():
            cg1 = QTreeWidgetItem(
                self.file_list_tree_widget,
                [
                    Path(path).name,
                    ' ',
                    metadata.axes,
                    str(metadata.shape),
                    str(metadata.dtype),
                    path,
                ],
            )
            cg1.setCheckState(1, Qt.Unchecked)



# flake8: noqa
import numpy
from scipy.ndimage import uniform_filter

from aydin.util.fast_uniform_filter.numba_cpu_uf import numba_cpu_uniform_filter
from aydin.util.fast_uniform_filter.numba_gpu_uf import numba_gpu_uniform_filter
from aydin.util.fast_uniform_filter.parallel_uf import parallel_uniform_filter
from aydin.io.datasets import fibsem, examples_single
from aydin.util.log.log import lsection, Log


def demo_par_uniform(image_name: str, image, size=128, repeats=32):
    Log.enable_output = True

    with lsection(f"size={size}:"):
        with lsection(f"Numba-CPU {image_name} (r={repeats}):"):
            for _ in range(repeats):
                numba_filtered_image = numba_cpu_uniform_filter(
                    image, size=size, mode="nearest"
                )

        with lsection(f"Parallel {image_name} (r={repeats}):"):
            for _ in range(repeats):
                parallel_filtered_image = parallel_uniform_filter(image, size=size)

        with lsection(f"Scipy {image_name} (r={repeats}):"):
            for _ in range(repeats):
                scipy_filtered_image = uniform_filter(image, size=size, mode="nearest")

        with lsection(f"Numba-CUDA {image_name} (r={repeats}):"):
            for _ in range(repeats):
                numba_cuda_filtered_image = numba_gpu_uniform_filter(image, size=size)

    # import napari
    #
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(
    #         numba_filtered_image, name='numba_filtered_image', colormap='plasma'
    #     )
    #     viewer.add_image(
    #         parallel_filtered_image, name='parallel_filtered_image', colormap='plasma'
    #     )
    #     viewer.add_image(
    #         scipy_filtered_image, name='scipy_filtered_image', colormap='plasma'
    #     )
    #
    #     viewer.add_image(
    #         numpy.abs(scipy_filtered_image - parallel_filtered_image),
    #         name='scipy_filtered_image - parallel_filtered_image',
    #         colormap='plasma',
    #     )
    #     viewer.add_image(
    #         numpy.abs(scipy_filtered_image - numba_filtered_image),
    #         name='scipy_filtered_image - numba_filtered_image',
    #         colormap='plasma',
    #     )

    numpy.testing.assert_array_almost_equal(
        parallel_filtered_image, scipy_filtered_image, decimal=3
    )

    numpy.testing.assert_array_almost_equal(
        numba_filtered_image, scipy_filtered_image, decimal=3
    )

    numpy.testing.assert_array_almost_equal(
        numba_cuda_filtered_image, scipy_filtered_image, decimal=3
    )


image_2d = fibsem(full=True).astype(numpy.float32)
image_3d = examples_single.royerlab_hcr.get_array().squeeze()[2]
image_4d = examples_single.hyman_hela.get_array().squeeze()

sizes = [3, 9, 64, 127, 317, 511]

for size in sizes:
    demo_par_uniform("fibsem 2D", image_2d, size=size)

for size in sizes:
    demo_par_uniform("islet 3D", image_3d, size=size, repeats=1)

for size in sizes:
    demo_par_uniform("hela 4D", image_4d, size=size, repeats=1)



import numpy

from aydin.features.base import FeatureGeneratorBase
from aydin.features.groups.base import FeatureGroupBase
from aydin.util.log.log import lprint, lsection


class ExtensibleFeatureGenerator(FeatureGeneratorBase):
    """
    Extensible Feature Generator class
    """

    def __init__(self):
        """
        Constructs an extensible feature generator
        """
        # Calls super constructor:
        super().__init__()

        # This list holds all the information for computing each 'group' of features:
        self.features_group_list = []

    def _load_internals(self, path: str):
        pass

    def add_feature_group(self, feature_group: FeatureGroupBase, *args, **kwargs):
        """
        Adds a feature to this feature generator.

        Parameters
        ----------
        feature_group : FeatureGroupBase
            feature group
        args
            additional arguments for function
        kwargs
            additional keyword arguments for function
        """
        self.features_group_list.append(feature_group)

    def clear_features(self):
        """
        Clears the features group list
        """
        self.features_group_list = []

    def get_num_features(self, ndim: int) -> int:
        """
        Returns the number of features when considering translations

        Parameters
        ----------
        ndim : int
            number of dimensions

        Returns
        -------
        nb_features : int
        """
        nb_features = 0
        for feature_group in self.features_group_list:
            nb_features += feature_group.num_features(ndim)
        return nb_features

    def get_receptive_field_radius(self) -> int:
        """
        Returns the receptive field radius in pixels

        Returns
        -------
        result : int
            receptive field radius in pixels
        """

        receptive_field_radius = 0
        for feature_group in self.features_group_list:
            receptive_field_radius = max(
                receptive_field_radius, feature_group.receptive_field_radius
            )
        return receptive_field_radius

    def compute(
        self,
        image,
        exclude_center_feature=False,
        exclude_center_value=False,
        features=None,
        feature_last_dim=True,
        passthrough_channels=None,
        num_reserved_features=0,
        excluded_voxels=None,
        spatial_feature_offset=None,
        spatial_feature_scale=None,
    ):
        """
        Computes the features given an image. If the input image is of shape (d,h,w),
        resulting features are of shape (n,d,h,w) where n is the number of features.

        Parameters
        ----------
        image : numpy.ndarray
        image for which features are computed
        exclude_center_feature : bool
        exclude_center_value : bool
        features
        feature_last_dim : bool
        passthrough_channels
        num_reserved_features : int
        excluded_voxels
        spatial_feature_offset
        spatial_feature_scale

        Returns
        -------
        feature array : numpy.ndarray

        """

        with lsection('Computing features'):

            # some important numbers:
            num_dims = len(image.shape)
            num_spatiotemp_dim = num_dims - 2
            num_batches = image.shape[0]
            num_channels = image.shape[1]
            num_features = self.get_num_features(num_spatiotemp_dim)

            # exclude_center_value can be a tuple, in that case each entry corresponds to a channel:
            if type(exclude_center_value) is not tuple:
                exclude_center_value = (exclude_center_value,) * num_channels

            # Fills in the default values for passthrough channels: False ==> by default channels are not passthrough.
            if passthrough_channels is None:
                passthrough_channels = (False,) * num_channels

            # Computes the number of features, taking into account: passthrough_channels, channels, and reserved_features:
            num_passthrough_channels = sum(1 if p else 0 for p in passthrough_channels)
            num_normal_features = num_features * (
                num_channels - num_passthrough_channels
            )
            num_total_features = (
                num_normal_features + num_passthrough_channels + num_reserved_features
            )

            # Creates feature array that will hold the final result:
            features = self.create_feature_array(image, num_total_features)

            # We iterate over batches:
            for batch_index in range(num_batches):

                with lsection(
                    f'Computing features for batch: {batch_index + 1}/{num_batches}'
                ):
                    feature_pointer = 0

                    # We iterate over channels:
                    for channel_index in range(num_channels):

                        with lsection(
                            f'Computing features for channel: {channel_index + 1}/{num_channels}'
                        ):
                            # We collect the 'exclude_center_value' for the current channel:
                            exclude_center_value_for_channel = exclude_center_value[
                                channel_index
                            ]

                            lprint(
                                f'Excluding center value for channel: {exclude_center_value_for_channel}'
                            )

                            # Image batch slice:
                            image_slice = (
                                batch_index,
                                channel_index,
                                *(slice(None),) * num_spatiotemp_dim,
                            )

                            lprint(f'Image slice: {image_slice}')

                            if passthrough_channels[channel_index]:
                                # A passthrough channel is simply fed directly as a feature:
                                lprint(
                                    f'Adding passthrough channel feature for channel index: {channel_index}'
                                )
                                batch_feature_slice = (
                                    slice(feature_pointer, feature_pointer + 1, 1),
                                    batch_index,
                                    *(slice(None),) * num_spatiotemp_dim,
                                )
                                features[batch_feature_slice] = image[
                                    image_slice
                                ].astype(self.dtype, copy=False)
                                feature_pointer += 1

                            else:
                                single_image = image[image_slice]

                                # Usefull code snippet for debugging features:
                                # with napari.gui_qt():
                                #      viewer = Viewer()
                                #      viewer.add_image(image_batch_gpu.get(), name='image')
                                #      viewer.add_image(rescale_intensity(image_integral_gpu.get(), in_range='image', out_range=(0, 1)), name='integral')

                                # Feature batch slice:
                                batch_feature_slice = (
                                    slice(None),
                                    batch_index,
                                    *(slice(None),) * num_spatiotemp_dim,
                                )
                                lprint(
                                    f'Feature slice for batch and channel: {batch_feature_slice}'
                                )

                                batch_channel_features = features[batch_feature_slice]

                                for feature_group in self.features_group_list:
                                    lprint(
                                        f'Computing feature {feature_group}, '  # , and kwargs={kwargs}'
                                    )

                                    # number of features in group:
                                    num_group_features = feature_group.num_features(
                                        num_spatiotemp_dim
                                    )

                                    # Excluded voxels:
                                    if excluded_voxels is None:
                                        excluded_voxels = []
                                    excluded_voxels_for_feature_group = []

                                    if exclude_center_value_for_channel:
                                        excluded_voxels_for_feature_group.extend(
                                            excluded_voxels
                                        )
                                        if (
                                            not (0,) * num_spatiotemp_dim
                                            in excluded_voxels_for_feature_group
                                        ):
                                            excluded_voxels_for_feature_group.append(
                                                (0,) * num_spatiotemp_dim
                                            )

                                    # We prepare feature generation for the group by setting the image,
                                    # any computation that can be factored should happen now:
                                    feature_group.prepare(
                                        single_image,
                                        excluded_voxels=excluded_voxels_for_feature_group,
                                        offset=spatial_feature_offset,
                                        scale=spatial_feature_scale,
                                    )

                                    for index in range(num_group_features):
                                        # feature array
                                        feature = batch_channel_features[
                                            feature_pointer
                                        ]

                                        # Computing the feature:
                                        feature_group.compute_feature(index, feature)

                                        # Increments feature index:
                                        feature_pointer += 1

                                    feature_group.finish()

            # # Usefull code snippet for debugging features:
            # import napari
            #
            # with napari.gui_qt():
            #     from napari import Viewer
            #
            #     viewer = Viewer()
            #     viewer.add_image(features, name='features')

            # 'collect_features_nD' puts the feature vector in axis 0.
            # The following line creates a view of the array
            # in which the features are indexed by the last dimension instead:
            if feature_last_dim:
                lprint('Move feature axis to the last axis ...')
                features = numpy.moveaxis(features, 0, -1)

            return features



import importlib

from aydin.restoration import denoise
from aydin.restoration.denoise.base import DenoiseRestorationBase


def get_denoiser_class_instance(variant, lower_level_args=None, it_transforms=None):
    """Returns instance of denoiser given a variant name, a set of arguments,
    and a list of transforms.

    Parameters
    ----------
    variant
        Variant to use
    lower_level_args
        Lower level args
    it_transforms
        Transforms to use.

    Returns
    -------
    Corresponding Restoration class
    """
    method_name_and_approach, implementation_name = variant.split("-")
    response = importlib.import_module(
        denoise.__name__ + '.' + method_name_and_approach.lower()
    )

    elem = [x for x in dir(response) if method_name_and_approach.lower() in x.lower()][
        0
    ]  # class name

    denoiser_class = response.__getattribute__(elem)

    return denoiser_class(
        lower_level_args=lower_level_args, it_transforms=it_transforms
    )


def get_list_of_denoiser_implementations():
    """
    Returns a list of denoiser implementations.
    Returns
    -------
    List of denoiser implementations.
    """
    denoiser_implementations = []
    descriptions = []

    for module in DenoiseRestorationBase.get_implementations_in_a_module(denoise):
        response = importlib.import_module(denoise.__name__ + '.' + module.name)

        elem = [x for x in dir(response) if module.name.replace('_', '') in x.lower()][
            0
        ]  # class name

        denoiser_class = response.__getattribute__(elem)
        denoiser_implementations += denoiser_class().implementations
        descriptions += denoiser_class().implementations_description

    return denoiser_implementations, descriptions



from abc import ABC

import numpy


class ShapeNormaliser(ABC):
    """Shape Normaliser

    Parameters
    ----------
    clip : bool
    epsilon : float
    shape_normalisation : bool
    transform : str

    """

    epsilon: float
    leave_as_float: bool
    clip: bool
    original_dtype: numpy.dtype

    def __init__(self, batch_axes=None, channel_axes=None):
        """Constructs a normalisers"""
        self.batch_axes = batch_axes
        self.channel_axes = channel_axes
        self.axis_permutation = None
        self.permutated_image_shape = None

    def normalise(self, array):
        """Normalises the given array in-place (if possible).

        Parameters
        ----------
        array : numpy.ndarray
            array to normalisers
        batch_dims : list
        channel_dims : list

        Returns
        -------
        array : numpy.ndarray

        """
        (
            array,
            self.axis_permutation,
            self.permutated_image_shape,
        ) = self.shape_normalize(
            array, batch_axes=self.batch_axes, channel_axes=self.channel_axes
        )

        return array

    def denormalise(self, array: numpy.ndarray, **kwargs):
        """Denormalises the given array in-place (if possible).

        Parameters
        ----------
        array : numpy.ndarray

        Returns
        -------
        array : numpy.ndarray

        """

        array = self.shape_denormalize(
            array,
            axes_permutation=self.axis_permutation,
            permutated_image_shape=self.permutated_image_shape,
        )

        return array

    @staticmethod
    def shape_normalize(image, batch_axes=None, channel_axes=None):
        """Permutates batch dimensions to the front and collapse into
        one dimension. Resulting array has to be in the form of (B,...)
        where B is the number of batch dimensions.

        Parameters
        ----------
        image : numpy.ndarray
        batch_axes : list
        channel_axes : list

        Returns
        -------
        Tuple of normalized_image, axes_permutation, permutated_image.shape : tuple

        """
        if batch_axes is None:
            batch_axes = (False,) * len(image.shape)
        if channel_axes is None:
            channel_axes = (False,) * len(image.shape)

        # Singleton dimensions are automatically batch dimension, unless it is a channel dimension, trivially.
        batch_axes = tuple(
            True if s == 1 and not c else b
            for b, c, s in zip(batch_axes, channel_axes, image.shape)
        )

        # get indices for different types of dimensions and their length
        batch_indices = [index for index, value in enumerate(batch_axes) if value]
        batch_length = int(numpy.prod([image.shape[index] for index in batch_indices]))

        channel_indices = (
            [index for index, value in enumerate(channel_axes) if value]
            if channel_axes
            else []
        )
        channel_length = int(
            numpy.prod([image.shape[index] for index in channel_indices])
        )

        spacetime_indices = [
            index
            for index in range(len(image.shape))
            if index not in batch_indices + channel_indices
        ]

        # Axes permutation
        axes_permutation = batch_indices + channel_indices + spacetime_indices

        # Bring all batch dimensions to front
        permutated_image = numpy.transpose(image, axes_permutation)

        # Collapse batch dimensions into one, same for channel dimensions
        spacetime_shape = tuple([image.shape[i] for i in spacetime_indices])
        normalized_shape = (batch_length, channel_length) + spacetime_shape
        normalized_shape = tuple(
            s for i, s in enumerate(normalized_shape) if s != 1 or i <= 1
        )

        # Reshape array:
        normalized_image = permutated_image.reshape(normalized_shape)

        return (normalized_image, axes_permutation, permutated_image.shape)

    @staticmethod
    def shape_denormalize(image, axes_permutation, permutated_image_shape):
        """Denormalizes the shape of an image from normalized form to the
        original image form.

        Parameters
        ----------
        image : numpy.ndarray
        axes_permutation : array_like
        permutated_image_shape : tuple

        Returns
        -------
        array : numpy.ndarray

        """
        spatiotemp_shape = image.shape[2:]

        # Number of spatio-temp dimensions:
        num_spatiotemp_dims = len(spatiotemp_shape)

        # If the input image has a different lengths along the spatio-temporal dimensions,
        # that's fine, we accommodate for it:
        # Note: that's only fine for spatio-temp dim, not batch or channels!
        adapted_permutated_image_shape = list(permutated_image_shape)
        adapted_permutated_image_shape[-num_spatiotemp_dims:] = spatiotemp_shape
        adapted_permutated_image_shape = tuple(adapted_permutated_image_shape)

        # Reshape the input to its permutated shape:
        permutated_image = image.reshape(adapted_permutated_image_shape)

        # Retrieves dimensions back:
        return numpy.transpose(permutated_image, axes=numpy.argsort(axes_permutation))



import math

import numpy
from copy import deepcopy
from scipy.ndimage.filters import convolve


def donut_filter(input_image):
    """
    Parameters
    ----------
    input_image

    Returns
    -------
    convolved image with kernel
    """
    k = [1] + [3 for _ in input_image.shape[1:-1]] + [1]
    kernel = numpy.ones(k) / 8
    k = [0] + [1 for _ in range(len(input_image.shape[1:-1]))] + [0]
    kernel[tuple(k)] = 0
    return convolve(input_image, kernel)


def train_image_generator(input_image, p=0.1):
    """
    Parameters
    ----------
    input_image
        input image
    p
        ratio of pixels being used for validation

    Returns
    -------
    training image
        p pixels were repalced by surrounding average, validation image: original image, mask: markers of validation
    """
    marker = numpy.random.uniform(size=input_image.shape)
    marker[marker > p] = 1
    marker[marker <= p] = 0
    marker = (1 - marker).astype(bool)

    img_filtered = donut_filter(input_image)
    img_train = deepcopy(input_image)
    img_train[marker] = img_filtered[marker]
    return img_train, input_image, marker


def val_data_generator(img_train, img_val, marker, batch_size, train_valid_ratio):
    """
    Parameters
    ----------
    img_train
    img_val
    marker
    batch_size
    train_valid_ratio
    """
    val_ind = numpy.random.randint(
        0, img_train.shape[0], math.ceil(img_train.shape[0] * train_valid_ratio)
    )
    img_train = deepcopy(img_train[val_ind])
    img_val = deepcopy(img_val[val_ind])
    marker = deepcopy(marker[val_ind])

    j = 0
    num_cycle = numpy.ceil(img_train.shape[0] / batch_size)
    while True:
        i = numpy.mod(j, num_cycle).astype(int)
        train_batch = img_train[batch_size * i : batch_size * (i + 1)]
        val_batch = img_val[batch_size * i : batch_size * (i + 1)]
        marker_batch = marker[batch_size * i : batch_size * (i + 1)]
        j += 1
        yield {
            'input': train_batch,
            'input_msk': marker_batch.astype(numpy.float32),
        }, val_batch


# may be useful in the future for generating validation data with less memory used.
# def val_ind_generator(image_train, p=0.1):
#     """
#
#     Parameters
#     ----------
#     image_train
#         input image
#     p
#         ratio of pixels being used for validation
#
#     Returns
#     -------
#     training image
#         p pixels were repalced by surrounding average, val_pix_values: original image, mask: markers of validation pixels
#
#     """
#     marker = numpy.random.randint(0, image_train.size, size=int(image_train.size * p))
#     marker = numpy.unravel_index(marker, image_train.shape)
#     val_pix_values = image_train[marker]
#     img_filtered = donut_filter(image_train)
#     image_train[marker] = img_filtered[marker]
#     return image_train, val_pix_values, marker



from math import sin, pi
from typing import Optional, Tuple

from aydin.features.groups.base import FeatureGroupBase
from aydin.util.log.log import lprint


class SpatialFeatures(FeatureGroupBase):
    def __init__(self, coarsening: int = 1, period: float = 0):
        super().__init__()

        self.coarsening = coarsening
        self.period = period

        self.image = None
        self.offset: Optional[Tuple[float, ...]] = None
        self.scale: Optional[Tuple[float, ...]] = None

    @property
    def receptive_field_radius(self) -> int:
        return 0

    def num_features(self, ndim: int) -> int:
        return ndim

    def prepare(
        self,
        image,
        offset: Optional[Tuple[float, ...]] = None,
        scale: Optional[Tuple[float, ...]] = None,
        **kwargs,
    ):
        self.image = image
        self.offset = offset
        self.scale = scale
        self.kwargs = kwargs

    def compute_feature(self, index: int, feature):
        lprint(
            f"Spatial feature {index}, offset={self.offset}, scale={self.scale}, coarsening={self.coarsening}, period={self.period} "
        )

        # Axis and dimensions:
        axis: int = index
        ndim = self.image.ndim

        # Scale and offset:
        offset = (0,) * ndim if self.offset is None else self.offset
        scale = (1,) * ndim if self.scale is None else self.scale

        # This is the offset for the corresponding axis:
        offset = offset[axis]
        scale = scale[axis]

        # Get the length on the dimension we want to iterate over
        dim_length = self.image.shape[axis]

        # Cache period and coarsening in local variables:
        coarsening = self.coarsening
        period = self.period

        # Iterate over the dimension of interest
        for index in range(dim_length):
            # Prepare a placeholder 'ndim' slice object
            slicing = [slice(None)] * len(feature.shape)

            # Change the slicing on the dimension we are iterating over
            slicing[axis] = slice(index, index + 1, 1)

            # Compute value and adds offset,
            # and we coarsen the resolution by a factor of two so that individual pixels cannot be identified:
            value = scale * float(coarsening * ((offset + index) // coarsening))

            if period != 0:
                value = sin((2 * pi * value) / period)

            # Put the normalized position value across dimensions we are NOT iterating over
            feature[tuple(slicing)] = value

    def finish(self):
        # Here we cleanup any resource allocated for the last feature computation:
        self.image = None



import math
from typing import Optional, Union, Tuple
import numpy
from pynndescent import NNDescent
from scipy.fft import dctn, idctn

from aydin.util.patch_size.patch_size import default_patch_size
from aydin.util.patch_transform.patch_transform import (
    extract_patches_nd,
    reconstruct_from_nd_patches,
)


def calibrate_denoise_bmnd(
    image, patch_size: Optional[Union[int, Tuple[int], str]] = None
):
    """
    Calibrates the BMnD denoiser for the given image and
    returns the optimal parameters obtained using the N2S
    loss.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate spectral denoiser for.

    patch_size: int
        Patch size for the 'image-to-patch' transform.
        Can be an int s that corresponds to isotropic
        patches of shape: (s,)*image.ndim, or a tuple
        of ints. By default (None) the patch size is
        chosen automatically to give the best results.


    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """

    # Default patch sizes vary with image dimension:
    patch_size = default_patch_size(image, patch_size, odd=True)

    best_parameters = {}

    # Memory needed:
    memory_needed = 2 * image.nbytes + 8 * image.nbytes * math.prod(patch_size)

    return denoise_bmnd, best_parameters, memory_needed


def denoise_bmnd(
    image,
    patch_size: Optional[Union[int, Tuple[int]]] = None,
    block_depth: Optional[int] = None,
    mode: str = 'median',
    reconstruction_gamma: float = 0,
    multi_core: bool = True,
):
    """
    Denoises the given image using the Block-Matching-nD approach.
    Our implementation follows the general outline of <a
    href="https://en.wikipedia.org/wiki/Block-matching_and_3D_filtering">
    Block-matching and 3D filtering (BM3D)</a> approaches.
    \n\n
    Note: This is currently only usable for small images, even for
    moderately sized images the computation time is prohibitive. We
    are planning to implement a GPU version to make it more usefull.


    Parameters
    ----------
    image: ArrayLike
        Image to denoise

    patch_size: int
        Patch size for the 'image-to-patch' transform.
        Can be: 'full' for a single patch covering the whole image, 'half', 'quarter',
        or an int s that corresponds to isotropic patches of shape: (s,)*image.ndim,
        or a tuple of ints. By default (None) the patch size is chosen automatically
        to give the best results.

    block_depth: int or None for default
        Block depth.

    mode: str
        Possible modes are: 'median', 'mean'.

    threshold: float
        Threshold between 0 and 1

    freq_bias_stength: float
        Frequency bias: closer to zero: no bias against high frequencies,
        closer to one and above: stronger bias towards high-frequencies.

    freq_cutoff: float
        Cutoff frequency, must be within [0, 1]. In addition

    order: float
        Filter order, typically an integer above 1.

    reconstruction_gamma: float
        Patch reconstruction parameter

    multi_core: bool
        By default we use as many cores as possible, in some cases, for small
        (test) images, it might be faster to run on a single core instead of
        starting the whole parallelization machinery.


    Returns
    -------
    Denoised image

    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # Normalise patch size:
    patch_size = default_patch_size(image, patch_size, odd=True)

    # Default value for block_depth:
    if block_depth is None:
        if type(patch_size) is int:
            block_depth = patch_size
        else:
            block_depth = max(patch_size)

    # First we apply the patch transform:
    patches = extract_patches_nd(image, patch_size=patch_size)

    axes = tuple(1 + a for a in range(image.ndim))
    patches = dctn(patches, axes=axes, workers=-1 if multi_core else 1)

    # reshape patches as vectors:
    original_patches_shape = patches.shape
    patches = patches.reshape(patches.shape[0], -1)

    # Build the data structure for nearest neighbor search:
    nn = NNDescent(patches, n_jobs=-1 if multi_core else 1)

    # for each patch we query for the nearest patch:
    indices, distances = nn.query(patches, k=block_depth)

    # reshape patches back to their original shape:
    patches = patches.reshape(original_patches_shape)

    # Extract blocks:
    blocks = patches[indices]

    # Denoise blocks:
    if mode == 'median':
        patches = numpy.median(blocks, axis=1)
    elif mode == 'mean':
        patches = numpy.mean(blocks, axis=1)
    else:
        raise ValueError(f"Unsupported mode: {mode}")
    del blocks

    patches = idctn(patches, axes=axes, workers=-1 if multi_core else 1)

    # Transform back from patches to image:
    denoised_image = reconstruct_from_nd_patches(
        patches, image.shape, gamma=reconstruction_gamma
    )

    # Cast back to float32 if needed:
    denoised_image = denoised_image.astype(numpy.float32, copy=False)

    return denoised_image



# flake8: noqa
import time

import numpy
from skimage.exposure import rescale_intensity

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io import io
from aydin.io.datasets import examples_single
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.nn import NNRegressor


def demo():
    image_path = examples_single.hyman_hela.get_path()
    image, metadata = io.imread(image_path)
    print(image.shape)
    # image = image[0:10, 15:35, 130:167, 130:177]
    image = image.astype(numpy.float16)
    print(image.shape)
    image = rescale_intensity(image, in_range='image', out_range=(0, 1))

    generator = StandardFeatureGenerator()

    regressor = NNRegressor()

    it = ImageTranslatorFGR(generator, regressor, balance_training_data=True)

    start = time.time()
    it.train(image, image)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    start = time.time()
    denoised = it.translate(image)
    stop = time.time()
    print(f"inference train: elapsed time:  {stop - start} ")

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(denoised, name='denoised')


demo()



# flake8: noqa
import time
from datetime import date
from os.path import join

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io import io
from aydin.io.datasets import examples_single
from aydin.io.folders import get_temp_folder
from aydin.io.io import mapped_tiff
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.lgbm import LGBMRegressor
from aydin.util.log.log import Log


def demo():
    Log.enable_output = True

    today = date.today()
    filename = f"result_{today.strftime('%m_%d_%Y_%H_%M_%S')}.tiff"

    image_path = examples_single.cognet_nanotube_100fps.get_path()
    array, metadata = io.imread(image_path)
    print(array.shape)
    train = array
    infer = array

    # print(f"Number of distinct features in image: {len(numpy.unique(infer))}")

    print(f"train: {train.shape}, inference:{infer.shape} ")

    batch_dims = (False, False, False)

    generator = StandardFeatureGenerator(
        kernel_widths=[5] + [1] + [3] * 10,
        kernel_scales=[1] + [2] + [2 ** i - 1 for i in range(2, 12)],
        kernel_shapes=['l2'] * 2 + ['l2'] * 3 + ['l1+nc'] * 5 + ['l1+oc'] * 2,
        max_level=9,
        include_spatial_features=True,
    )

    regressor = LGBMRegressor(patience=20, gpu_prediction=True)

    it = ImageTranslatorFGR(generator, regressor)

    start = time.time()
    it.train(train, train, batch_axes=batch_dims)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    output_file = join(get_temp_folder(), filename)

    print(f"Output file: {output_file}")

    # We write the stack to a temp file:
    with mapped_tiff(output_file, infer.shape, infer.dtype) as denoised_tiff:
        # denoised = offcore_array(whole.shape, whole.dtype)

        start = time.time()
        denoised = it.translate(
            infer, translated_image=denoised_tiff, batch_axes=batch_dims, tile_size=512
        )
        stop = time.time()

        print(f"Writing to file: {output_file} ")
        denoised_tiff[...] = denoised[...]

        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(train, name='train')
            viewer.add_image(infer, name='infer')
            print(f"Inference: elapsed time:  {stop - start} ")
            viewer.add_image(denoised, name='denoised')


demo()



import numpy

from aydin.io.datasets import camera
from aydin.it.transforms.histogram import HistogramEqualisationTransform


def test_histogram():
    image = camera()

    ht = HistogramEqualisationTransform()

    preprocessed = ht.preprocess(image)
    postprocessed = ht.postprocess(preprocessed)

    # import napari
    #
    # viewer = napari.Viewer()
    # viewer.add_image(image, name='image')
    # viewer.add_image(preprocessed, name='preprocessed')
    # viewer.add_image(postprocessed, name='postprocessed')
    # napari.run()

    assert postprocessed.dtype == image.dtype
    assert postprocessed.shape == image.shape
    assert numpy.abs(postprocessed - image).mean() < 1e-3



# flake8: noqa
import numpy
from scipy.ndimage import gaussian_filter

from aydin.util.denoise_nd.denoise_nd import extend_nd


def test_denoise_nd():
    # raw function that only supports 2D images:
    def function(image, sigma):
        if image.ndim != 2:
            raise RuntimeError("Function only supports arrays of dimensions 2")
        return gaussian_filter(image, sigma)

    # extended function that supports all dimension (with all caveats associated to how we actually do this extension...)
    @extend_nd(available_dims=[2])
    def extended_function(image, sigma):
        return function(image, sigma)

    # Wrongly extended function: we pretend that it can do dim 1 when in fact it can't!
    @extend_nd(available_dims=[1, 2])
    def wrongly_extended_function(image, sigma):
        return function(image, sigma)

    image = numpy.zeros((32,))
    image[16] = 1

    try:
        function(image, sigma=1)
        assert False
    except RuntimeError as e:
        # expected!
        assert True

    try:
        extended_function(image, sigma=1)
        assert True
    except RuntimeError as e:
        assert False

    try:
        wrongly_extended_function(image, sigma=1)
        assert False
    except RuntimeError as e:
        assert True

    image = numpy.zeros((32, 5, 64))
    image[16, 2, 32] = 1

    try:
        function(image, sigma=1)
        assert False
    except RuntimeError as e:
        # expected!
        assert True

    try:
        extended_function(image, sigma=1)
        assert True
    except RuntimeError as e:
        assert False



from aydin.io.datasets import examples_single
from aydin.it.transforms.fixedpattern import FixedPatternTransform
from aydin.util.log.log import Log


def demo_fixed_pattern_real():
    Log.override_test_exclusion = True

    image = examples_single.huang_fixed_pattern_noise.get_array()  # [:, 0:64, 0:64]

    bs = FixedPatternTransform()  # axes=[1, 2])

    pre_processed = bs.preprocess(image)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(pre_processed, name='pre_processed')


demo_fixed_pattern_real()



from typing import Union, Sequence

import numpy
from numba import jit

# from numpy.typing import ArrayLike

from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lsection, lprint


class AttenuationTransform(ImageTransformBase):
    """Axis-aligned attenuation correction

    Corrects intensity attenuation of an image along a given list of axis.
    This is usefull to correct for signal attenuation over time or along
    space. Currently only linear attenuation is supported. More modes on the
    way.
    """

    preprocess_description = (
        "Suppresses axis-aligned attenuation"
        + ImageTransformBase.preprocess_description
    )
    postprocess_description = (
        "Reapplies attenuation" + ImageTransformBase.postprocess_description
    )
    postprocess_supported = True
    postprocess_recommended = False

    def __init__(
        self,
        axes: Union[None, int, Sequence[int]] = None,
        mode: str = 'linear',
        priority: float = 0.321,
        **kwargs,
    ):
        """
        Constructs a Attenuation Corrector

        Parameters
        ----------
        axes : Union[None, int, Sequence[int]]
            Axis or list of axes over which to correct attenuation.
            If None the axes are automatically determined.

        mode : str
            Attenuation fitting mode, only currently supported: 'linear'

        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)
        self.axis = (
            ([tuple(ac) for ac in axes] if type(axes) is not int else (axes,))
            if axes is not None
            else None
        )
        self.mode = mode
        self._original_dtype = None
        self._corrections = {}
        self._axis = None

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_original_dtype']
        del state['_corrections']
        del state['_axis']
        return state

    def __str__(self):
        return f'{type(self).__name__} (mode={self.mode})'

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):

        with lsection(
            f"Correcting attenuation for array of shape: {array.shape} and dtype: {array.dtype}:"
        ):

            self._axis = self.axis
            if self._axis is None:
                self._axis = list(range(0, array.ndim))

            if type(self._axis) == int:
                self._axis = [self._axis]

            self._original_dtype = array.dtype

            # Allocate new array to store result:
            new_array = array.astype(dtype=numpy.float32, copy=True)

            # Reset cache for undoing transforms:
            self._corrections = {}

            # Iterating over dimensions to correct:
            for axis in self._axis:
                lprint(f"Correcting along axis: {axis}")
                if array.shape[axis] <= 1:
                    continue

                # Which axis do we need to compute the medians over?
                axis_to_compute_medians = tuple(
                    i for i in range(0, array.ndim) if i != axis
                )
                medians_along_axis = numpy.median(
                    array, axis=axis_to_compute_medians
                ).astype(numpy.float32, copy=True)

                # Compute trend:
                trend = self._trend_fit(medians_along_axis)

                # Make the trend array ndimensional:
                trend_nd_slice_tuple = tuple(
                    None if i != axis else slice(None) for i in range(0, array.ndim)
                )
                trend_nd = trend[trend_nd_slice_tuple]

                # Compute correction:
                trend_max = numpy.max(trend)
                correction = trend_max / trend_nd
                correction = correction.astype(dtype=numpy.float32, copy=True)

                # fast in-place multiply:
                _in_place_multiply(new_array, correction)

                # remember correction to be able to undo it:
                self._corrections[axis] = correction

            return new_array

    def _trend_fit(self, medians_along_axis):
        if self.mode == 'linear':
            # prepare data for Thiel-Sen fitting:
            X = numpy.arange(0, len(medians_along_axis), 1).reshape(-1, 1)
            y = medians_along_axis
            # Thiel-Sen fitting:
            from sklearn.linear_model import TheilSenRegressor

            reg = TheilSenRegressor(random_state=0).fit(X, y)
            # Compute the estimated trend:
            trend = reg.predict(X)
            return trend
        else:
            raise ValueError(f"Mode '{self.mode}' for attenuation correction unknown!")

    def postprocess(self, array):

        if not self.do_postprocess:
            return array

        with lsection(
            f"Reapplying attenuation for array of shape: {array.shape} and dtype: {array.dtype}:"
        ):

            # Allocate new array to store result:
            new_array = array.astype(dtype=numpy.float32, copy=True)

            for axis in reversed(self._axis):
                lprint(f"Correcting along axis: {axis}")
                correction = self._corrections[axis]
                inverse_correction = 1.0 / correction
                inverse_correction = inverse_correction.astype(dtype=correction.dtype)
                _in_place_multiply(new_array, inverse_correction)

            new_array = new_array.astype(self._original_dtype, copy=False)

            return new_array


@jit(
    nopython=True,
    parallel=False,
    error_model='numpy',
    fastmath={'contract', 'afn', 'reassoc'},
)
def _in_place_multiply(image, factor):
    image *= factor



import gc
import math
import os
from abc import ABC, abstractmethod
from os.path import join
from typing import Union, List, Optional, Tuple
import jsonpickle
import numpy
import psutil

from aydin.analysis.blind_spot_analysis import auto_detect_blindspots
from aydin.it.exceptions.base import ArrayShapeDoesNotMatchError
from aydin.it.normalisers.shape import ShapeNormaliser
from aydin.it.transforms.base import ImageTransformBase
from aydin.util.array.nd import nd_split_slices, remove_margin_slice
from aydin.util.misc.json import encode_indent
from aydin.util.log.log import lprint, lsection
from aydin.util.offcore.offcore import offcore_array


class ImageTranslatorBase(ABC):
    """Image Translator base class

    Notes
    -----

    """

    def __init__(
        self,
        monitor=None,
        blind_spots: Optional[Union[str, List[Tuple[int]]]] = None,
        tile_min_margin: int = 8,
        tile_max_margin: Optional[int] = None,
        max_memory_usage_ratio: float = 0.9,
        max_tiling_overhead: float = 0.1,
    ):
        """Image Translator base class

        Parameters
        ----------
        monitor
        blind_spots : Optional[Union[str, List[Tuple[int]]]]
        tile_min_margin : int
        tile_max_margin : Optional[int]
        max_memory_usage_ratio : float
            Maximum allowed memory load.
        max_tiling_overhead : float
            Maximum allowed margin overhead during tiling.

        """
        self.self_supervised = False
        self.monitor = monitor
        self.blind_spots = blind_spots
        self.tile_max_margin = tile_max_margin
        self.tile_min_margin = tile_min_margin

        self.transforms_list: List[ImageTransformBase] = []

        self.max_memory_usage_ratio = max_memory_usage_ratio
        self.max_tiling_overhead = max_tiling_overhead
        self.max_voxels_per_tile = 768 ** 3

        self.callback_period = 3
        self.last_callback_time_sec = -math.inf

        self.loss_history = None

    def add_transform(self, transform: ImageTransformBase, sort: bool = True):
        """Adds the given transform to the self.transforms_list

        Parameters
        ----------
        transform : ImageTransformBase

        """
        self.transforms_list.append(transform)
        if sort:
            self.transforms_list = sorted(
                self.transforms_list, key=lambda t: t.priority
            )

    def clear_transforms(self):
        """Clears the transforms list"""
        self.transforms_list.clear()

    def _transform_preprocess_image(self, image):
        with lsection("transform preprocess:"):
            for transform in self.transforms_list:
                lprint(f"applying transform: {transform}")
                try:
                    image = transform.preprocess(image)
                except BaseException as e:
                    error_message = str(e).replace('\n', ', ')
                    lprint(
                        f"Preprocessing failed for {transform} with: {error_message} "
                    )
                    import traceback
                    import sys

                    traceback.print_exception(*sys.exc_info())

        return image

    def _transform_postprocess_image(self, image):
        with lsection("transform postprocess"):
            for transform in reversed(self.transforms_list):
                lprint(f"applying transform: {transform}")
                try:
                    image = transform.postprocess(image)
                except BaseException as e:
                    error_message = str(e).replace('\n', ', ')
                    lprint(
                        f"Postprocessing failed for {transform} with: {error_message}"
                    )
                    import traceback
                    import sys

                    traceback.print_exception(*sys.exc_info())

        return image

    @abstractmethod
    def _train(
        self, input_image, target_image, train_valid_ratio, callback_period, jinv
    ):
        """This function supposed to take normalized input image only

        Parameters
        ----------
        input_image
        target_image
        train_valid_ratio
        callback_period
        jinv

        Returns
        -------

        """
        raise NotImplementedError()

    def stop_training(self):
        pass

    @abstractmethod
    def _translate(self, input_image, image_slice=None, whole_image_shape=None):
        """Translates an input image into an output image according to the learned function

        Parameters
        ----------
        input_image
            input image
        image_slice
        whole_image_shape

        Returns
        -------

        """
        raise NotImplementedError()

    @abstractmethod
    def _load_internals(self, path: str):
        raise NotImplementedError()

    def _estimate_memory_needed_and_available(self, image):
        """

        Parameters
        ----------
        image

        Returns
        -------

        """
        # TODO: make sure if this makes sense with GPU too
        # By default there is no memory needed and all memory is available
        return 0, psutil.virtual_memory().total

    def train(
        self,
        input_image,
        target_image=None,
        batch_axes=None,
        channel_axes=None,
        train_valid_ratio=0.1,
        callback_period=3,
        jinv=None,
    ):
        """Train to translate a given input image to a given output image.
        This has a lot of the machinery for batching and more...
        """

        if target_image is None:
            target_image = input_image

        with lsection(
            f"Learning to translate from image of dimensions {str(input_image.shape)} to {str(target_image.shape)}, batch_axes={batch_axes}, channel_axes={channel_axes}, jinv={jinv}."
        ):

            lprint('Running garbage collector...')
            gc.collect()

            # If we use the same image for input and output then we are in a self-supervised setting:
            self.self_supervised = input_image is target_image

            if self.self_supervised:
                lprint('Training is self-supervised.')
            else:
                lprint('Training is supervised.')

            # Let's apply the transforms:
            input_image = self._transform_preprocess_image(input_image)
            target_image = (
                input_image
                if self.self_supervised
                else self._transform_preprocess_image(target_image)
            )

            if batch_axes is None:  # set default batch_axes value:
                batch_axes = (False,) * len(input_image.shape)
            if channel_axes is None:
                channel_axes = (False,) * len(input_image.shape)

            # parse batch and channel dim args
            batch_axes, channel_axes = self.parse_axes_args(
                batch_axes, channel_axes, len(input_image.shape)
            )

            if len(batch_axes) != len(input_image.shape):
                raise ArrayShapeDoesNotMatchError(
                    'The length of batch_dims and input_image dimensions are different.'
                )

            # Shape normalisers to use:
            shape_normaliser = ShapeNormaliser(
                batch_axes=batch_axes, channel_axes=channel_axes
            )

            # Axis normalisation:
            shape_normalised_input_image = shape_normaliser.normalise(input_image)
            shape_normalised_target_image = (
                shape_normalised_input_image
                if self.self_supervised
                else shape_normaliser.normalise(target_image)
            )

            # Automatic blind-spot discovery:
            if self.blind_spots == 'auto':
                lprint(
                    "Automatic discovery of noise autocorrelation and specification of N2S blind-spots activated!"
                )
                self.blind_spots = auto_detect_blindspots(
                    shape_normalised_input_image[0, 0]
                )[0]
                lprint(f"Blind spots: {self.blind_spots}")

            # Verify that input and target images have same shape:
            # We do this after normalisation because that's easier
            # we only compare the spatio-temporal dimensions ( not batches or channels)
            if (
                shape_normalised_input_image.shape[2:]
                != shape_normalised_target_image.shape[2:]
            ):
                raise ArrayShapeDoesNotMatchError(
                    'Input and Output image shape does not match!'
                )

            self._train(
                shape_normalised_input_image,
                shape_normalised_target_image,
                train_valid_ratio=train_valid_ratio,
                callback_period=callback_period,
                jinv=jinv,
            )

    def translate(
        self,
        input_image,
        translated_image=None,
        batch_axes=None,
        channel_axes=None,
        tile_size=None,
    ):

        with lsection(
            f"Predicting output image from input image of dimension {input_image.shape}, batch_axes={batch_axes}, channel_axes={channel_axes}"
        ):

            # Let's apply the transforms:
            input_image = self._transform_preprocess_image(input_image)

            # set default batch_axes and channel_axes values:
            if batch_axes is None:
                batch_axes = (False,) * len(input_image.shape)
            if channel_axes is None:
                channel_axes = (False,) * len(input_image.shape)

            # parse batch and chan dim args
            batch_axes, channel_axes = self.parse_axes_args(
                batch_axes, channel_axes, len(input_image.shape)
            )

            if len(batch_axes) != len(
                input_image.shape
            ):  # Sanity check when using not-default batch dims:
                raise ArrayShapeDoesNotMatchError(
                    'batch_dims does not have same number of dimensions with input_image!'
                )

            # Number of spatio-temporal dimensions:
            num_spatiotemp_dim = sum(
                0 if b or c else 1 for b, c in zip(batch_axes, channel_axes)
            )

            # Shape normalisers to use:
            shape_normaliser = ShapeNormaliser(
                batch_axes=batch_axes, channel_axes=channel_axes
            )

            # First we normalise the input values:
            shape_normalised_input_image = shape_normaliser.normalise(input_image)

            # Spatio-temporal shape:
            spatiotemp_shape = shape_normalised_input_image.shape[-num_spatiotemp_dim:]

            shape_normalised_translated_image = None

            if tile_size == 0:
                # we _force_ no tilling, this is _not_ the default.

                # We translate:
                shape_normalised_translated_image = self._translate(
                    shape_normalised_input_image,
                    whole_image_shape=shape_normalised_input_image.shape,
                )

            else:

                # We do need to do tiled inference because of a lack of memory
                # or because a small batch size was requested:

                normalised_input_shape = shape_normalised_input_image.shape

                # We get the tilling strategy:
                # tile_size, shape, min_margin, max_margin
                tilling_strategy, margins = self._get_tilling_strategy_and_margins(
                    shape_normalised_input_image,
                    self.max_voxels_per_tile,
                    self.tile_min_margin,
                    self.tile_max_margin,
                    suggested_tile_size=tile_size,
                )
                lprint(f"Tilling strategy: {tilling_strategy}")
                lprint(f"Margins for tiles: {margins} .")

                # tile slice objects (with and without margins):
                tile_slices_margins = list(
                    nd_split_slices(
                        normalised_input_shape, tilling_strategy, margins=margins
                    )
                )
                tile_slices = list(
                    nd_split_slices(normalised_input_shape, tilling_strategy)
                )

                # Number of tiles:
                number_of_tiles = len(tile_slices)
                lprint(f"Number of tiles (slices): {number_of_tiles}")

                # We create slice list:
                slicezip = zip(tile_slices_margins, tile_slices)

                counter = 1
                for slice_margin_tuple, slice_tuple in slicezip:
                    with lsection(
                        f"Current tile: {counter}/{number_of_tiles}, slice: {slice_tuple} "
                    ):

                        # We first extract the tile image:
                        input_image_tile = shape_normalised_input_image[
                            slice_margin_tuple
                        ].copy()

                        # We do the actual translation:
                        lprint("Translating...")
                        translated_image_tile = self._translate(
                            input_image_tile,
                            image_slice=slice_margin_tuple,
                            whole_image_shape=shape_normalised_input_image.shape,
                        )

                        # We compute the slice needed to cut out the margins:
                        lprint("Removing margins...")
                        remove_margin_slice_tuple = remove_margin_slice(
                            normalised_input_shape, slice_margin_tuple, slice_tuple
                        )

                        # We allocate -just in time- the translated array if needed:
                        # if the array is already provided, it must of course have the right dimensions...
                        if shape_normalised_translated_image is None:
                            translated_image_shape = (
                                shape_normalised_input_image.shape[:2]
                                + spatiotemp_shape
                            )
                            shape_normalised_translated_image = offcore_array(
                                shape=translated_image_shape,
                                dtype=translated_image_tile.dtype,
                                max_memory_usage_ratio=self.max_memory_usage_ratio,
                            )

                        # We plug in the batch without margins into the destination image:
                        lprint("Inserting translated batch into result image...")
                        shape_normalised_translated_image[
                            slice_tuple
                        ] = translated_image_tile[remove_margin_slice_tuple]

                        counter += 1

            # Then we shape denormalise:
            shape_denormalised_translated_image = shape_normaliser.denormalise(
                shape_normalised_translated_image
            )

            # Let's undo the transforms:
            shape_denormalised_translated_image = self._transform_postprocess_image(
                shape_denormalised_translated_image
            )

            if translated_image is None:
                translated_image = shape_denormalised_translated_image
            else:
                translated_image[...] = shape_denormalised_translated_image

            return translated_image

    def _get_tilling_strategy_and_margins(
        self,
        image,
        max_voxels_per_tile,
        min_margin,
        max_margin,
        suggested_tile_size=None,
    ):

        # We will store the batch strategy as a list of integers representing the number of chunks per dimension:
        with lsection("Determine tilling strategy:"):

            suggested_tile_size = (
                math.inf if suggested_tile_size is None else suggested_tile_size
            )

            # image shape:
            shape = image.shape
            num_spatio_temp_dim = num_spatiotemp_dim = len(shape) - 2

            lprint(f"image shape             = {shape}")
            lprint(f"max_voxels_per_tile     = {max_voxels_per_tile}")

            # Estimated amount of memory needed for storing all features:
            (
                estimated_memory_needed,
                total_memory_available,
            ) = self._estimate_memory_needed_and_available(image)
            lprint(f"Estimated amount of memory needed: {estimated_memory_needed}")

            # Available physical memory :
            total_memory_available *= self.max_memory_usage_ratio

            lprint(
                f"Available memory (we reserve 10% for 'comfort'): {total_memory_available}"
            )

            # How much do we need to tile because of memory, if at all?
            split_factor_mem = estimated_memory_needed / total_memory_available
            lprint(
                f"How much do we need to tile because of memory? : {split_factor_mem} times."
            )

            # how much do we have to tile because of the limit on the number of voxels per tile?
            split_factor_max_voxels = image.size / max_voxels_per_tile
            lprint(
                f"How much do we need to tile because of the limit on the number of voxels per tile? : {split_factor_max_voxels} times."
            )

            # how much do we have to tile because of the suggested tile size?
            split_factor_suggested_tile_size = image.size / (
                suggested_tile_size ** num_spatio_temp_dim
            )
            lprint(
                f"How much do we need to tile because of the suggested tile size? : {split_factor_suggested_tile_size} times."
            )

            # we keep the max:
            desired_split_factor = max(
                split_factor_mem,
                split_factor_max_voxels,
                split_factor_suggested_tile_size,
            )
            # We cannot split less than 1 time:
            desired_split_factor = max(1, int(math.ceil(desired_split_factor)))
            lprint(f"Desired split factor: {desired_split_factor}")

            # Number of batches:
            num_batches = shape[0]

            # Does the number of batches split the data enough?
            if num_batches < desired_split_factor:
                # Not enough splitting happening along the batch dimension, we need to split further:
                # how much?
                rest_split_factor = desired_split_factor / num_batches
                lprint(
                    f"Not enough splitting happening along the batch dimension, we need to split spatio-temp dims by: {rest_split_factor}"
                )

                # let's split the dimensions in a way proportional to their lengths:
                split_per_dim = (rest_split_factor / numpy.prod(shape[2:])) ** (
                    1 / num_spatio_temp_dim
                )
                lprint(f"Splitting per dimension: {split_per_dim}")

                # We split proportionally to each axis but do not exceed the rest_split_factor per axis:
                spatiotemp_tilling_strategy = tuple(
                    max(1, min(rest_split_factor, int(math.ceil(split_per_dim * s))))
                    for s in shape[2:]
                )

                tilling_strategy = (num_batches, 1) + spatiotemp_tilling_strategy
                lprint(f"Preliminary tilling strategy is: {tilling_strategy}")

                # We correct for eventual oversplitting by favouring splitting over the front dimensions:
                current_splitting_factor = 1
                corrected_tilling_strategy = []
                split_factor_reached = False
                for i, s in enumerate(tilling_strategy):

                    if split_factor_reached:
                        corrected_tilling_strategy.append(1)
                    else:
                        corrected_tilling_strategy.append(s)
                        current_splitting_factor *= s

                    if current_splitting_factor >= desired_split_factor:
                        split_factor_reached = True

                tilling_strategy = tuple(corrected_tilling_strategy)

            else:
                tilling_strategy = (desired_split_factor, 1) + tuple(
                    1 for s in shape[2:]
                )

            lprint(f"Tilling strategy is: {tilling_strategy}")

            # Handles defaults:
            if max_margin is None:
                max_margin = math.inf
            if min_margin is None:
                min_margin = 0

            # First we estimate the shape of a tile:

            estimated_tile_shape = tuple(
                int(round(s / ts)) for s, ts in zip(shape[2:], tilling_strategy[2:])
            )
            lprint(f"The estimated tile shape is: {estimated_tile_shape}")

            # Limit margins:
            # We automatically set the margin of the tile size:
            # the max-margin factor guarantees that tilling will incur no more than a given max tiling overhead:
            margin_factor = 0.5 * (
                ((1 + self.max_tiling_overhead) ** (1 / num_spatiotemp_dim)) - 1
            )
            margins = tuple(int(s * margin_factor) for s in estimated_tile_shape)

            # Limit the margin to something reasonable (provided or automatically computed):
            margins = tuple(min(max_margin, m) for m in margins)
            margins = tuple(max(min_margin, m) for m in margins)

            # We add the batch and channel dimensions:
            margins = (0, 0) + margins

            # We only need margins if we split a dimension:
            margins = tuple(
                (0 if split == 1 else margin)
                for margin, split in zip(margins, tilling_strategy)
            )

            return tilling_strategy, margins

    def save(self, path: str):
        """
        Saves a 'all-batteries-included' image translation model at a given path (folder).

        Parameters
        ----------
        path : str
            path to save to

        Returns
        -------

        """
        os.makedirs(path, exist_ok=True)

        frozen = encode_indent(self)

        lprint(f"Saving image translator to: {path}")
        with open(join(path, "image_translation.json"), "w") as json_file:
            json_file.write(frozen)

        return frozen

    @staticmethod
    def load(path: str):
        """
        Returns an 'all-batteries-included' image translation model at a given path (folder).

        Parameters
        ----------
        path : str
            path to load from.

        Returns
        -------

        """
        lprint(f"Loading image translator from: {path}")
        with open(join(path, "image_translation.json"), "r") as json_file:
            frozen = json_file.read()

        thawed = jsonpickle.decode(frozen)

        thawed._load_internals(path)

        return thawed

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['input_normaliser']
        del state['target_normaliser']
        return state

    @staticmethod
    def parse_axes_args(
        batch_axes: Union[List[int], List[bool]],
        chan_axes: Union[List[int], List[bool]],
        ndim: int,
    ):
        """


        Parameters
        ----------
        batch_axes : Union[List[int], List[bool]]
        chan_axes : Union[List[int], List[bool]]
        ndim : int

        Returns
        -------

        """
        if (
            batch_axes == []
            or batch_axes is None
            or all(isinstance(x, bool) for x in batch_axes)
        ) and (
            chan_axes == []
            or chan_axes is None
            or all(isinstance(x, bool) for x in chan_axes)
        ):
            # check if it is all boolean values then check if it is correct size then return
            batch_result, chan_result = batch_axes, chan_axes
        elif (
            batch_axes == []
            or batch_axes is None
            or all(isinstance(x, int) for x in batch_axes)
        ) and (
            chan_axes == []
            or chan_axes is None
            or all(isinstance(x, int) for x in chan_axes)
        ):
            # check if it is list of indices and can make meaningful boolean arrays, if so return
            if any(i < 0 or i >= ndim for i in batch_axes) or any(
                i < 0 or i >= ndim for i in chan_axes
            ):
                raise Exception(
                    "No axes index can be smaller than zero or bigger than ndim-1!"
                )

            batch_result = [True if i in batch_axes else False for i in range(ndim)]
            chan_result = [True if i in chan_axes else False for i in range(ndim)]
        else:
            raise Exception(
                "Axes arguments has to be boolean arrays or integer index arrays!"
            )

        ndim_spacetime = ndim - (batch_result.count(True) + chan_result.count(True))
        if ndim_spacetime > 4 or ndim_spacetime < 1:
            print(batch_result)
            print(chan_result)
            print(
                ndim, batch_result.count(True), chan_result.count(True), ndim_spacetime
            )
            raise Exception(
                "Number of spacetime axes cannot be more than 4 and cannot be less than 1!"
            )

        if any(
            [batch_result[i] is True and chan_result[i] is True for i in range(ndim)]
        ):
            raise Exception("No axes can be both batch and chan axes!")

        return batch_result, chan_result



import torch
from torch.optim import Adam


class ESAdam(Adam):
    r"""Implements a modifified version of the Adam algorithm that adds noise to the
    the .

    """

    def __init__(self, params, start_noise_level=0.001, **kwargs):
        super().__init__(params, **kwargs)

        self.start_noise_level = start_noise_level
        self.step_counter = 0

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = super().step(closure)

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad: torch.Tensor = p.grad.data
                if grad.is_sparse:
                    continue

                step_size = group['lr']

                p.data += (
                    step_size
                    * (self.start_noise_level / (1 + self.step_counter))
                    * (torch.rand_like(p.data) - 0.5)
                )

        self.step_counter += 1

        return loss



import sys
from urllib.request import urlretrieve
from pathlib import Path
import os
import zipfile


def resource_path(relative_path):
    """
    Method to get resource path from bundle

    Parameters
    ----------
    relative_path : str

    Returns
    -------
    resource path : str

    """
    try:
        base_path = sys._MEIPASS

        return os.path.join(base_path, os.path.basename(relative_path))
    except Exception:
        base_path = os.path.abspath("..")

    return os.path.join(base_path, relative_path)


def download_and_extract_zipresource(url, targetdir='.'):
    """
    Method to download and extract a zipresource from a url.

    Parameters
    ----------
    url : str
    targetdir : str

    """

    # Check if target directory exists, if not create
    targetdir = Path(targetdir)
    if not targetdir.is_dir():
        targetdir.mkdir(parents=True, exist_ok=True)

    # Compute relative path to resource
    relative_path_to_zip = str(targetdir) + '/' + os.path.basename(url)
    print("relativepath= ", relative_path_to_zip)

    # Check if target resource already exists, retrieve the resource if not exists
    if os.path.exists(relative_path_to_zip[:-4]):
        print("Resource already exists, nothing to download")
    else:
        urlretrieve(url, relative_path_to_zip)
        # Extract the content
        with zipfile.ZipFile(relative_path_to_zip, "r") as zip_ref:
            zip_ref.extractall(str(targetdir))

        # Delete zip file
        Path.unlink(Path(relative_path_to_zip))



import numpy

# from numpy.typing import ArrayLike

from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lsection, lprint


class DeskewTransform(ImageTransformBase):
    """(Integral) Stack Deskewer

    Denoising is more effective if voxels carrying correlated signal are close to each other. When a stack is skewed
    -- as resulting in some imaging modalities -- correlated voxels that should be close in space are far from each
    other. Thus, deskewing the image before denoising is highly recommended. Importantly, the deskewing must be
    'integral', meaning that it must not interpolate voxel values, which is a unadvised lossy operation. Integral
    stack deskewing consists in applying an integral shear transformation to a stack. Two axes need to be specified:
    the 'z'-axis and the 'skew'-axis along which shifting happens. The delta parameter controls the amount of shift
    per plane - must be an integer. We automatically snap the delta value to the closest integer. Padding is supported.

    Note: this only works for images with at least 3 dimensions. Does nothing
    on images with less than 3 dimensions.
    """

    preprocess_description = "Deskew image" + ImageTransformBase.preprocess_description
    postprocess_description = (
        "Reskew image" + ImageTransformBase.postprocess_description
    )
    postprocess_supported = True
    postprocess_recommended = True

    def __init__(
        self,
        delta: float = 0,
        z_axis: int = 0,
        skew_axis: int = 1,
        pad: bool = True,
        priority: float = 0.4,
        **kwargs,
    ):
        """
        Constructs a stack deskewer

        Parameters
        ----------
        delta : float
            How much shifting from one plane to the next
        z_axis : int
            Axis for which the amount of shift depends upon.
        skew_axis : int
            Axis over which the image is shifted.
        pad : bool
            True for padding before rolling, this is useful
            because normal padding is rarely enough.
        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)
        self.delta = int(round(delta))
        self.z_axis = z_axis
        self.skew_axis = skew_axis
        self.pad = pad

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        # nothing to exclude
        return state

    def __str__(self):
        return (
            f'{type(self).__name__} (delta={self.delta},'
            f' z_axis={self.z_axis},'
            f' skew_axis={self.skew_axis},'
            f' pad={self.pad} )'
        )

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):
        with lsection(
            f"Deskewing (delta={self.delta}, z_axis={self.z_axis}, skew_axis={self.skew_axis}, pad={self.pad}) array of shape: {array.shape} and dtype: {array.dtype}:"
        ):
            if array.ndim >= 3:
                return self.deskew(array)
            else:
                return array

    def postprocess(self, array):
        if not self.do_postprocess:
            return array
        with lsection(
            f"Undoing deskew for array of shape: {array.shape} and dtype: {array.dtype}:"
        ):
            if array.ndim >= 3:
                return self.reskew(array)
            else:
                return array

    def deskew(self, array, pad_mode='wrap'):
        array = self._permutate(array)
        array = self._skew_transform(
            array, self.delta, pad=True, crop=False, pad_mode=pad_mode
        )
        array = self._depermutate(array)
        return array

    def reskew(self, array):
        array = self._permutate(array)
        array = self._skew_transform(
            array, -self.delta, pad=False, crop=True, pad_mode=''
        )
        array = self._depermutate(array)
        return array

    def _permutate(self, array):
        permutation = self._get_permutation(array)
        array = numpy.transpose(array, axes=permutation)
        return array

    def _depermutate(self, array):
        permutation = self._get_permutation(array, inverse=True)
        array = numpy.transpose(array, axes=permutation)
        return array

    def _get_permutation(self, array, inverse=False):
        permutation = (self.z_axis, self.skew_axis) + tuple(
            axis
            for axis in range(array.ndim)
            if axis not in [self.z_axis, self.skew_axis]
        )
        if inverse:
            permutation = numpy.argsort(permutation)
        return permutation

    @staticmethod
    def _skew_transform(array, delta, pad, crop, pad_mode='wrap'):
        """
        This method assumes that the first dimension (index=0) is the z dimension,
        and the second dimension (index=1) is the 'skewed' dimension.
        The array can have arbitrary dimensions after that...
        We also assume that the array has been properly padded so that we can 'roll' the
        skewed dimension without fear or regret.
        """

        num_z_planes = array.shape[0]
        pad_length = abs(delta * num_z_planes)

        if pad:
            padding = (pad_length, 0) if delta < 0 else (0, pad_length)
            array = numpy.pad(
                array,
                pad_width=((0, 0), padding) + ((0, 0),) * (array.ndim - 2),
                mode=pad_mode,
            )
        else:
            array = array.copy()

        for zi in range(num_z_planes):
            array[zi, ...] = numpy.roll(array[zi, ...], shift=delta * zi, axis=0)

        if crop:
            cropping = (
                slice(pad_length, None, 1) if delta > 0 else slice(0, -pad_length, 1)
            )
            crop_slice = (slice(None), cropping) + (slice(None),) * (array.ndim - 2)
            array = array[crop_slice]

        return array



from math import ceil
from typing import Tuple, Optional

import numba
import numpy
from numba import jit, prange

__fastmath = {'contract', 'afn', 'reassoc'}
__error_model = 'numpy'


def fast_shift(
    image, shift: Tuple[int], output=None, add=False, parallelism: Optional[int] = 8
):

    # Save original image dtype:
    original_dtype = image.dtype

    # Numba does not support float16 yet:
    dtype = numpy.float32 if original_dtype == numpy.float16 else original_dtype
    image = image.astype(dtype=dtype, copy=False)

    # return type:
    output = numpy.zeros_like(image) if output is None else output

    parallelism = numba.get_num_threads() if parallelism is None else parallelism

    if image.ndim == 1:
        output = _fast_shift_1d(image, shift, output, add, parallelism)
    elif image.ndim == 2:
        output = _fast_shift_2d(image, shift, output, add, parallelism)
    elif image.ndim == 3:
        output = _fast_shift_3d(image, shift, output, add, parallelism)
    elif image.ndim == 4:
        output = _fast_shift_4d(image, shift, output, add, parallelism)
    else:
        raise ValueError("Image dimensions must be less or equal to 4.")

    # Make sure that output image has correct dtype:
    output = output.astype(dtype=original_dtype, copy=False)

    return output


@jit(nopython=True, parallel=True, error_model=__error_model, fastmath=__fastmath)
def _fast_shift_1d(image, shift: Tuple[int], output, add: bool, parallelism: int):

    width = image.shape[0]
    chunk_width = int(ceil(width / parallelism))
    shift_x = shift[0]

    for c in prange(parallelism):
        for k in range(chunk_width):
            src_x = k + c * chunk_width
            dst_x = src_x + shift_x
            if 0 <= dst_x < width and src_x < width:
                if add:
                    output[dst_x] += image[src_x]
                else:
                    output[dst_x] = image[src_x]

    return output


@jit(nopython=True, parallel=True, error_model=__error_model, fastmath=__fastmath)
def _fast_shift_2d(image, shift: Tuple[int], output, add: bool, parallelism: int):

    height = image.shape[0]
    width = image.shape[1]
    chunk_height = int(ceil(height / parallelism))
    shift_y = shift[0]
    shift_x = shift[1]

    for c in prange(parallelism):
        for k in range(chunk_height):
            src_y = k + c * chunk_height
            dst_y = src_y + shift_y
            if 0 <= dst_y < height and src_y < height:
                for src_x in range(width):
                    dst_x = src_x + shift_x
                    if 0 <= dst_x < width:
                        if add:
                            output[dst_y, dst_x] += image[src_y, src_x]
                        else:
                            output[dst_y, dst_x] = image[src_y, src_x]
    return output


@jit(nopython=True, parallel=True, error_model=__error_model, fastmath=__fastmath)
def _fast_shift_3d(image, shift: Tuple[int], output, add: bool, parallelism: int):

    depth = image.shape[0]
    height = image.shape[1]
    width = image.shape[2]
    shift_z = shift[0]
    shift_y = shift[1]
    shift_x = shift[2]
    chunk_depth = int(ceil(depth / parallelism))

    for c in prange(parallelism):
        for k in range(chunk_depth):
            src_z = k + c * chunk_depth
            dst_z = src_z + shift_z
            if 0 <= dst_z < depth and src_z < depth:
                for src_y in range(height):
                    dst_y = src_y + shift_y
                    if 0 <= dst_y < height:
                        for src_x in range(width):
                            dst_x = src_x + shift_x
                            if 0 <= dst_x < width:
                                if add:
                                    output[dst_z, dst_y, dst_x] += image[
                                        src_z, src_y, src_x
                                    ]
                                else:
                                    output[dst_z, dst_y, dst_x] = image[
                                        src_z, src_y, src_x
                                    ]
    return output


@jit(nopython=True, parallel=True, error_model=__error_model, fastmath=__fastmath)
def _fast_shift_4d(image, shift: Tuple[int], output, add: bool, parallelism: int):

    length = image.shape[0]
    depth = image.shape[1]
    height = image.shape[2]
    width = image.shape[3]
    shift_w = shift[0]
    shift_z = shift[1]
    shift_y = shift[2]
    shift_x = shift[3]
    chunk_length = int(ceil(length / parallelism))

    for c in prange(parallelism):
        for k in range(chunk_length):
            src_w = k + c * chunk_length
            dst_w = src_w + shift_w
            if 0 <= dst_w < length and src_w < length:
                for src_z in range(depth):
                    dst_z = src_z + shift_z
                    if 0 <= dst_z < depth:
                        for src_y in range(height):
                            dst_y = src_y + shift_y
                            if 0 <= dst_y < height:
                                for src_x in range(width):
                                    dst_x = src_x + shift_x
                                    if 0 <= dst_x < width:
                                        if add:
                                            output[dst_w, dst_z, dst_y, dst_x] += image[
                                                src_w, src_z, src_y, src_x
                                            ]
                                        else:
                                            output[dst_w, dst_z, dst_y, dst_x] = image[
                                                src_w, src_z, src_y, src_x
                                            ]
    return output



import pytest
from numpy.random.mtrand import normal

from aydin.analysis.snr_estimate import snr_estimate
from aydin.io.datasets import camera, normalise
from aydin.util.log.log import lprint, Log


def demo_snr_estimate(display: bool = False):
    Log.enable_output = True

    clean_image = normalise(camera())

    noise = normal(size=clean_image.size).reshape(*clean_image.shape)

    noisy_image_1 = clean_image + 100 * noise
    noisy_image_2 = clean_image + 10 * noise
    noisy_image_3 = clean_image + noise
    noisy_image_4 = clean_image + 0.1 * noise
    noisy_image_5 = clean_image + 0.01 * noise

    noise1_dB = snr_estimate(noisy_image_1)
    noise2_dB = snr_estimate(noisy_image_2)
    noise3_dB = snr_estimate(noisy_image_3)
    noise4_dB = snr_estimate(noisy_image_4)
    noise5_dB = snr_estimate(noisy_image_5)

    lprint(f"noise1_dB={noise1_dB}")
    lprint(f"noise2_dB={noise2_dB}")
    lprint(f"noise3_dB={noise3_dB}")
    lprint(f"noise4_dB={noise4_dB}")
    lprint(f"noise5_dB={noise5_dB}")

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(clean_image, name='clean_image')
            viewer.add_image(noisy_image_1, name='noisy_image_1')
            viewer.add_image(noisy_image_2, name='noisy_image_2')
            viewer.add_image(noisy_image_3, name='noisy_image_3')
            viewer.add_image(noisy_image_4, name='noisy_image_4')
            viewer.add_image(noisy_image_5, name='noisy_image_5')

    assert noise1_dB == pytest.approx(-42, 1, 3)
    assert noise2_dB == pytest.approx(-20, 1, 1)
    assert noise3_dB == pytest.approx(-0, 1, 1)
    assert noise4_dB == pytest.approx(19, 1, 1)
    assert noise5_dB == pytest.approx(33, 1, 1)


demo_snr_estimate()



import numpy
from skimage.util import random_noise

from aydin.io.datasets import normalise, camera
from aydin.it.transforms.highpass import HighpassTransform


def test_high_pass():
    image = normalise(camera())
    image = random_noise(image, mode="s&p", amount=0.03, seed=0, clip=False)

    ac = HighpassTransform()

    preprocessed = ac.preprocess(image)
    postprocessed = ac.postprocess(preprocessed)

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(preprocessed, name='preprocessed')
    #     viewer.add_image(postprocessed, name='postprocessed')

    assert image.shape == postprocessed.shape
    assert image.dtype == postprocessed.dtype

    assert postprocessed.dtype == image.dtype
    assert numpy.abs(postprocessed - image).mean() < 1e-8



import gc
import math
import multiprocessing
import shutil
from os.path import join
from tempfile import mkdtemp
from typing import Sequence
import numpy
from catboost import CatBoostRegressor, CatBoostError, Pool

from aydin.regression.base import RegressorBase
from aydin.regression.cb_utils.callbacks import CatBoostStopTrainingCallback
from aydin.util.log.log import lsection, lprint


class CBRegressor(RegressorBase):
    """CatBoost Regressor."""

    model: CatBoostRegressor

    def __init__(
        self,
        num_leaves: int = 512,
        max_num_estimators: int = 2048,
        min_num_estimators: int = 512,
        max_bin: int = None,
        learning_rate: float = 0.01,
        loss: str = 'l1',
        patience: int = 32,
        compute_load: float = 0.95,
        gpu: bool = True,
        gpu_devices: Sequence[int] = None,
    ):
        """Constructs a CatBoost regressor.

        Parameters
        ----------
        num_leaves : int
            Number of leaves.
        max_num_estimators : int
            Maximum number of estimators
        min_num_estimators : int
            Minimum number of estimators
        max_bin : int
            Maximum number of allowed bins
        learning_rate : float
            Learning rate for the catboost model
        loss : str
            Type of loss to be used
        patience : int
            Number of rounds required for early stopping
        compute_load : float
            Allowed load on computational resources in percentage
        gpu : bool
            Flag to tell catboost try to use GPU or do not
        gpu_devices : Sequence[int]
            List of GPU device indices to be used by CatBoost
        """
        super().__init__()

        self.force_verbose_eval = False
        self.stop_training_callback = CatBoostStopTrainingCallback()

        self.num_leaves = num_leaves
        self.max_num_estimators = max_num_estimators
        self.min_num_estimators = min_num_estimators
        if max_bin is None:
            self.max_bin = 254 if gpu else 512
        else:
            self.max_bin = max_bin
        self.learning_rate = learning_rate
        self.metric = loss
        self.early_stopping_rounds = patience
        self.compute_load = compute_load

        self.gpu = gpu
        self.gpu_devices = gpu_devices

        with lsection("CB Regressor"):
            lprint(f"patience: {self.early_stopping_rounds}")
            lprint(f"gpu: {self.gpu}")

    def recommended_max_num_datapoints(self) -> int:
        """Recommended maximum number of datapoints

        Returns
        -------
        int

        """
        return 40e6 if self.gpu else 1e6

    def _get_params(
        self, num_samples, num_features, learning_rate, dtype, use_gpu, train_folder
    ):

        min_data_in_leaf = 20 + int(0.01 * (num_samples / self.num_leaves))
        # lprint(f'min_data_in_leaf: {min_data_in_leaf}')

        objective = self.metric
        if objective == 'l1':
            objective = 'MAE'
        elif objective == 'l2':
            objective = 'RMSE'

        max_depth = max(3, int(math.log2(self.num_leaves)) - 1)
        max_depth = min(max_depth, 8) if use_gpu else max_depth

        gpu_ram_type = 'CpuPinnedMemory' if num_samples > 10e6 else 'GpuRam'

        params = {
            "iterations": self.max_num_estimators,
            "task_type": "GPU" if use_gpu else "CPU",
            "devices": 'NULL'
            if self.gpu_devices is None
            else ':'.join(self.gpu_devices),  # uses all available GPUs
            'objective': objective,
            "loss_function": self.metric.upper(),
            "allow_writing_files": True,
            "train_dir": train_folder,
            "max_bin": self.max_bin,
            "rsm": None if use_gpu else 0.8,  # same as GBM
            "thread_count": max(
                1, int(self.compute_load * multiprocessing.cpu_count())
            ),
            "gpu_cat_features_storage": gpu_ram_type,
            'max_depth': max_depth,
            'early_stopping_rounds': self.early_stopping_rounds,
            'bagging_temperature': 1,
            'min_data_in_leaf': min_data_in_leaf,
            'l2_leaf_reg': 30,
            'feature_border_type': 'UniformAndQuantiles'
            # "num_leaves": self.num_leaves,
        }

        params["learning_rate"] = learning_rate

        return params

    def stop_fit(self):
        self.stop_training_callback.continue_training = False

    def _fit(
        self, x_train, y_train, x_valid=None, y_valid=None, regressor_callback=None
    ):

        with lsection("CatBoost regressor fitting:"):

            nb_data_points = y_train.shape[0]
            self.num_features = x_train.shape[-1]
            has_valid_dataset = x_valid is not None and y_valid is not None

            lprint(f"Number of data points: {nb_data_points}")
            if has_valid_dataset:
                lprint(f"Number of validation data points: {y_valid.shape[0]}")
            lprint(f"Number of features per data point: {self.num_features}")

            # Train folder to store training info:
            train_folder = mkdtemp(prefix="catboost_training_")

            self.__epoch_counter = 0

            model = None

            with lsection(
                f"CatBoost regressor fitting now using {f'GPU({self.gpu_devices})' if self.gpu else 'CPU'} "
            ):
                # CatBoost prefers float32 arrays:
                x_train = x_train.astype(numpy.float32, copy=False)
                y_train = y_train.astype(numpy.float32, copy=False)

                xy_train_pool = Pool(data=x_train, label=y_train)

                # Keep this for later:
                x_train_shape = x_train.shape
                y_train_shape = y_train.shape
                x_train_dtype = x_train.dtype

                # Give a chance to reclaim this memory if needed:
                x_train, y_train = None, None

                # CatBoost fails (best_iter == 0 or too small) sometimes to train if learning rate is too high, this loops
                # tries increasingly smaller learning rates until training succeeds (best_iter>min_n_estimators)
                learning_rate = self.learning_rate
                for i in range(10):
                    if not self.stop_training_callback.continue_training:
                        break
                    lprint(f"Trying learning rate of {learning_rate}")

                    # The purpose of this try block is to protect against failure to use GPU.
                    try:
                        params = self._get_params(
                            num_samples=nb_data_points,
                            num_features=self.num_features,
                            learning_rate=learning_rate,
                            dtype=x_train_dtype,
                            use_gpu=self.gpu,
                            train_folder=train_folder,
                        )
                        lprint(f"Initialising CatBoost with {params}")
                        model = CatBoostRegressor(**params)

                        lprint(
                            f"Fitting CatBoost model for: X{x_train_shape} -> y{y_train_shape}"
                        )
                        model.fit(
                            X=xy_train_pool,
                            eval_set=(x_valid, y_valid) if has_valid_dataset else None,
                            early_stopping_rounds=self.early_stopping_rounds,
                            use_best_model=has_valid_dataset,
                            # callbacks=[self.stop_training_callback],
                        )
                    except CatBoostError as e:
                        print(e)
                        lprint("GPU training likely failed, switching to CPU.")
                        self.gpu = False
                        # next attempt next...
                        continue

                    # Training succeeds when the best iteration is not the zeroth's iteration.
                    # best_iteration_ might be None if there is no validation data provided...
                    if (
                        model.best_iteration_ is None
                        or model.best_iteration_ > self.min_num_estimators
                    ):
                        self.learning_rate = learning_rate
                        lprint(
                            f"CatBoost fitting succeeded! new learning rate for regressor: {learning_rate}"
                        )
                        break
                    else:
                        # Reduce learning rate:
                        if learning_rate is None:
                            learning_rate = 0.01
                        learning_rate *= 0.5
                        lprint(
                            f"CatBoost fitting failed! best_iteration=={model.best_iteration_} < {self.min_num_estimators} reducing learning rate to: {learning_rate}"
                        )
                        gc.collect()

                lprint("CatBoost fitting done.")

            if has_valid_dataset and model is not None:
                valid_loss = model.get_best_score()['validation'][params['objective']]
                self.last_valid_loss = valid_loss

            loss_history = _read_loss_history(train_folder)
            if (
                'catboost_training_' in train_folder
            ):  # sanity check as we delete a lot of files!
                shutil.rmtree(train_folder, ignore_errors=True)

            gc.collect()
            return _CBModel(model, loss_history)


def _read_loss_history(train_folder):
    training_loss = numpy.genfromtxt(
        join(train_folder, "learn_error.tsv"), delimiter="\t", skip_header=1
    )[:, 1]
    validation_loss = numpy.genfromtxt(
        join(train_folder, "test_error.tsv"), delimiter="\t", skip_header=1
    )[:, 1]
    return {'training': training_loss, 'validation': validation_loss}


class _CBModel:
    def __init__(self, model, loss_history):
        self.model: CatBoostRegressor = model
        self.loss_history = loss_history

    def _save_internals(self, path: str):
        if self.model is not None:
            cb_model_file = join(path, 'catboost_model.txt')
            self.model.save_model(cb_model_file)

    def _load_internals(self, path: str):
        cb_model_file = join(path, 'catboost_model.txt')
        self.model = CatBoostRegressor()
        self.model.load_model(cb_model_file)

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['model']
        return state

    def predict(self, x):

        with lsection("CatBoost regressor prediction"):

            lprint(f"Number of data points             : {x.shape[0]}")
            lprint(f"Number of features per data points: {x.shape[-1]}")

            lprint("Converting input to CatBoost's Pool format...")
            # CatBoost prefers float32 arrays:
            x = x.astype(dtype=numpy.float32, copy=False)
            # Create pool object:
            x_pool = Pool(data=x)

            with lsection("CatBoost prediction now"):
                prediction = self.model.predict(
                    x_pool, thread_count=-1, verbose=True
                ).astype(numpy.float32, copy=False)
                # task_type='CPU') # YOUHOUU!

            lprint("CatBoost regressor predicting done!")
            return prediction



from sklearn.linear_model import LinearRegression, HuberRegressor, Lasso

from aydin.regression.base import RegressorBase
from aydin.util.log.log import lsection, lprint


class LinearRegressor(RegressorBase):
    """Linear Regressor.
    \n\n
    Note: Fast but overall poor performance -- as expected.
    """

    def __init__(
        self,
        mode: str = 'huber',
        max_num_iterations: int = 512,
        alpha: float = 1,
        beta: float = 0.0001,
        **kwargs,
    ):
        """Constructs a linear regressor

        Parameters
        ----------
        mode : str
            Regression mode, supported options: 'lasso', 'huber', and 'linear'

        max_num_iterations: int
            Maximum number of iterations.

        alpha: float
            Regularisation weight for L1 sparsity term for Lasso regression
            (mode='lasso').

        beta: float
            Regularisation weight for Huber regression (mode='huber').


        kwargs
        """
        super().__init__()
        self.mode = mode
        self.max_num_iterations = max_num_iterations
        self.alpha = alpha
        self.beta = beta

        with lsection("Linear Regressor"):
            lprint(f"mode : {self.mode}")
            lprint(f"alpha: {self.alpha}")

    def _fit(
        self, x_train, y_train, x_valid=None, y_valid=None, regressor_callback=None
    ):
        """Fits function y=f(x) given training pairs (x_train, y_train).
        Stops when performance stops improving on the test dataset: (x_test, y_test).
        """
        if self.mode == 'lasso':
            model = Lasso(alpha=self.alpha, max_iter=self.max_num_iterations)
        elif self.mode == 'huber':
            model = HuberRegressor(max_iter=self.max_num_iterations, alpha=self.beta)
        elif self.mode == 'linear':
            model = LinearRegression(n_jobs=-1)
        else:
            raise Exception(f"Unknown mode: {self.mode}!")

        model.fit(x_train, y_train)

        return _LinearModel(model)


class _LinearModel:
    def __init__(self, model):
        self.model = model
        self.loss_history = {'training': [], 'validation': []}

    def _save_internals(self, path: str):
        pass

    def _load_internals(self, path: str):
        pass

    def predict(self, x):

        with lsection("Linear regressor prediction"):

            lprint(f"Number of data points             : {x.shape[0]}")
            lprint(f"Number of features per data points: {x.shape[-1]}")

            with lsection("Linear prediction now"):
                prediction = self.model.predict(x)

            lprint("Linear regressor predicting done!")
            return prediction



import numpy
from numba import jit

# from numpy.typing import ArrayLike
from sklearn.base import TransformerMixin
from sklearn.preprocessing import PowerTransformer, QuantileTransformer

from aydin.it.transforms.base import ImageTransformBase
from aydin.util.log.log import lsection, lprint


class VarianceStabilisationTransform(ImageTransformBase):
    """Variance Stabilization Transform (VST)
    and other range compression functions.

    Applies either the Yeo-Johnson, Box-Cox, Anscomb or some other VST
    transform to the image. Variance stabilization turns an image with
    Poisson or in general non-Gaussian noise into an image with approximately
    Gaussian noise, which often facilitates denoising.
    Also, in general for images with very skewed histograms with for example
    dark regions of large extent and only a few very bright regions,
    it is often beneficial to apply a sqrt or log transformation before
    denoising.
    """

    preprocess_description = (
        "Apply transform" + ImageTransformBase.preprocess_description
    )
    postprocess_description = (
        "Undo transform" + ImageTransformBase.postprocess_description
    )
    postprocess_supported = True
    postprocess_recommended = True

    def __init__(
        self,
        mode: str = 'anscomb',
        leave_as_float: bool = True,
        priority: float = 0.11,
        **kwargs,
    ):

        """
        Constructs a Variance Stabilisation Transform

        Parameters
        ----------
        mode : str
            Variance stabilisation mode: 'yeo-johnson', 'box-cox',
            'quantile', 'anscomb', 'log', 'sqrt' and 'identity'.
            Our tests show that Anscomb seems to work best for
            typical microscopy images. However, your mileage may
            vary. Images with strong contrast and very little
            noise might not need variance stabilisation, but could
            still benefit from value compression (log or sqrt).
            The advantage of the Yeo-Johnson and Box-Cox transforms
            is that they are adaptive, the transform is adjusted
            to match the input distribution.

        leave_as_float : bool
            Does not attempt to cast back to original non-flloat data
            type (8, 16, 32 bit integer), but instead leaves as 32 bit float.

        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.
        """
        super().__init__(priority=priority, **kwargs)

        self.mode = mode
        self.leave_as_float = leave_as_float

        self._original_dtype = None
        self._min = None
        self._max = None
        self._transform: TransformerMixin = None

        lprint(f"Instanciating: {self}")

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_original_dtype']
        del state['_min']
        del state['_max']
        del state['_transform']
        return state

    def __str__(self):
        return f'{type(self).__name__}' f' (mode={self.mode})'

    def __repr__(self):
        return self.__str__()

    def preprocess(self, array):

        with lsection(
            f"Stabilising variance ({self.mode}) for array of shape: {array.shape} and dtype: {array.dtype}"
        ):
            # Let's ensure we are working with floats:
            self._original_dtype = array.dtype
            array = array.astype(numpy.float32, copy=False)

            if self.mode == 'identity':
                new_array = array.copy()
            elif self.mode == 'sqrt':
                min_value = array.min()
                new_array = _f_sqrt(array, min_value)
                self._min_value = min_value
            elif self.mode == 'log':
                min_value = array.min()
                new_array = _f_log(array, min_value)
                self._min_value = min_value
            elif self.mode == 'anscomb':
                min_value = array.min()
                new_array = _f_anscomb(array, min_value)
                self._min_value = min_value
            elif self.mode == 'yeo-johnson' or self.mode == 'box-cox':

                # If the image has negative values, or too small variance, we fallback to yeo-johnson:
                mode = (
                    'yeo-johnson'
                    if array.min() < 0 or numpy.var(numpy.log1p(array.ravel())) < 1e-6
                    else self.mode
                )
                lprint(f"Actual mode used: {mode} ")

                # We prepare array to make it compatible to sklearn API and save shape:
                shape = array.shape
                array = array.ravel()[:, numpy.newaxis]

                try:
                    # Instantiate sklearn power transform:
                    power_transform = PowerTransformer(method=mode, standardize=True)

                    numpy.seterr(all='raise')
                    # Fit tranform:
                    power_transform.fit(array)
                except (Warning, FloatingPointError):
                    lprint(
                        f"VST {mode} failed for some numerical reasons, falling back to yeo-johnson."
                    )
                    mode = 'yeo-johnson'

                    # Instantiate sklearn power transform:
                    power_transform = PowerTransformer(method=mode, standardize=True)

                    # Fit tranform:
                    power_transform.fit(array)

                # Apply transform:
                new_array = power_transform.transform(array)

                # Reshape array:
                new_array = numpy.reshape(new_array, newshape=shape)

                # Save transform:
                self._transform = power_transform
            elif self.mode == 'quantile':

                # We prepare array to make it compatible to sklearn API and save shape:
                shape = array.shape
                array = array.ravel()[:, numpy.newaxis]

                # Instantiate sklearn quantile transform:
                power_transform = QuantileTransformer(
                    n_quantiles=1024, output_distribution='normal'
                )

                # Fit tranform:
                power_transform.fit(array)

                # Apply transform:
                new_array = power_transform.transform(array)

                # Reshape array:
                new_array = numpy.reshape(new_array, newshape=shape)

                # Save transform:
                self._transform = power_transform
            else:
                raise ValueError(f"Unsupported or incompatible VST mode: {self.mode}")

            return new_array

    def postprocess(self, array):

        if not self.do_postprocess:
            return array

        with lsection(
            f"Applying inverse variance stabilisation transform ({self.mode}) for array of shape: {array.shape} and dtype: {array.dtype}"
        ):

            if self.mode == 'identity':
                new_array = array.copy()
            elif self.mode == 'sqrt':
                new_array = _i_sqrt(array, self._min_value)
            elif self.mode == 'log':
                new_array = _i_log(array, self._min_value)
            elif self.mode == 'anscomb':
                new_array = _i_anscomb(array, self._min_value)
            elif (
                self.mode == 'yeo-johnson'
                or self.mode == 'box-cox'
                or self.mode == 'quantile'
            ):

                # We prepare array to make it compatible to sklearn API and save shape:
                shape = array.shape
                array = array.ravel()[:, numpy.newaxis]

                # Apply inverse transform:
                new_array = self._transform.inverse_transform(array)

                # Reshape array:
                new_array = numpy.reshape(new_array, newshape=shape)

                # reset transform:
                self._transform = None
            else:
                raise ValueError(f"Unsupported VST mode: {self.mode}")

            if not self.leave_as_float:
                # convert back to original dtype:
                new_array = new_array.astype(dtype=self._original_dtype, copy=False)

            return new_array


@jit(nopython=True, parallel=True, error_model='numpy')
def _f_sqrt(image, min_value):
    return numpy.sqrt(image - min_value)


@jit(nopython=True, parallel=True, error_model='numpy')
def _i_sqrt(image, min_value):
    return image ** _F(2) + min_value


@jit(nopython=True, parallel=True, error_model='numpy')
def _f_log(image, min_value):
    return numpy.log1p(image - min_value)


@jit(nopython=True, parallel=True, error_model='numpy')
def _i_log(image, min_value):
    return numpy.exp(image) - _F(1) + min_value


@jit(nopython=True, parallel=True, error_model='numpy')
def _f_anscomb(image, min_value):
    return _F(2) * numpy.sqrt(image - min_value + _F(0.375))


@jit(nopython=True, parallel=True, error_model='numpy')
def _i_anscomb(image, min_value):
    return (image / _F(2)) ** _F(2) + min_value - _F(0.375)


@jit(nopython=True, error_model='numpy')
def _F(x):
    return numpy.asarray(x, dtype=numpy.float32)



# flake8: noqa
import numpy
import matplotlib.pyplot as plt
from aydin.it.cnn.models.unet_2d import UNet2DModel


def bbox_idx(x, thresh=None):
    if thresh is None:
        thresh = 0
    return [[numpy.amin(i), numpy.amax(i)] for i in numpy.where(x != thresh)]


# Check the receptive field of masking approach ==========================================
def demo_masking():
    # Setup a model
    batch_norm = None
    supervised = False
    shiftconv = False  # True
    batch_size = 1
    num_lyr = 5
    input_image = numpy.zeros((1, 512, 512, 1))
    input_shape = input_image.shape
    input_image1 = numpy.copy(input_image)
    input_image1[0, int(input_shape[1] / 2), int(input_shape[2] / 2), 0] = 1
    model = UNet2DModel(
        input_shape[1:],
        rot_batch_size=1,
        num_lyr=num_lyr,
        normalization=batch_norm,
        supervised=supervised,
        shiftconv=shiftconv,
        res_node=1,
    )
    # run the model with a 0 image
    if not shiftconv and not supervised:
        pred0 = model.predict(
            [input_image, numpy.ones(input_shape)], batch_size=batch_size, verbose=1
        )
        pred1 = model.predict(
            [input_image1, numpy.ones(input_shape)], batch_size=batch_size, verbose=1
        )
    else:
        pred0 = model.predict(input_image, batch_size=batch_size, verbose=1)
        pred1 = model.predict(input_image1, batch_size=batch_size, verbose=1)

    # calculate receptive field
    receptive_field_idx = numpy.array(bbox_idx(pred1))
    receptive_field_size = receptive_field_idx[:, -1] - receptive_field_idx[:, 0] + 1
    print(f'rf size: {receptive_field_size}')

    # Visualize receptive field (by passing only 1 pixel info)
    pred01 = numpy.copy(pred1)
    pred01[pred01 == 0] = numpy.amin(pred1)
    plt.figure()
    plt.subplot(1, 2, 1)
    plt.imshow(pred1.squeeze())
    plt.axis('off')
    plt.subplot(1, 2, 2)
    plt.imshow(pred01.squeeze())
    plt.axis('off')
    plt.title('Masking approach')
    plt.show()


demo_masking()


# Check the receptive field of shiftconv approach ==========================================
def demo_shiftconv():
    batch_norm = None
    supervised = False
    shiftconv = True
    batch_size = 1
    num_lyr = 5
    input_image = numpy.zeros((1, 512, 512, 1))
    input_shape = input_image.shape
    input_image1 = numpy.copy(input_image)
    input_image1[0, int(input_shape[1] / 2), int(input_shape[2] / 2), 0] = 1
    model = UNet2DModel(
        input_shape[1:],
        rot_batch_size=1,
        num_lyr=num_lyr,
        normalization=batch_norm,
        supervised=supervised,
        shiftconv=shiftconv,
        res_node=1,
    )
    # run the model with a 0 image
    if not shiftconv and not supervised:
        pred0 = model.predict(
            [input_image, numpy.ones(input_shape)], batch_size=batch_size, verbose=1
        )
        pred1 = model.predict(
            [input_image1, numpy.ones(input_shape)], batch_size=batch_size, verbose=1
        )
    else:
        pred0 = model.predict(input_image, batch_size=batch_size, verbose=1)
        pred1 = model.predict(input_image1, batch_size=batch_size, verbose=1)

    # calculate receptive field
    receptive_field_idx = numpy.array(bbox_idx(pred1))
    receptive_field_size = receptive_field_idx[:, -1] - receptive_field_idx[:, 0] + 1
    print(f'rf size: {receptive_field_size}')

    # Visualize receptive field (by passing only 1 pixel info)
    pred01 = numpy.copy(pred1)
    pred01[pred01 == 0] = numpy.amin(pred1)
    plt.figure()
    plt.subplot(1, 2, 1)
    plt.imshow(pred1.squeeze())
    plt.axis('off')
    plt.subplot(1, 2, 2)
    plt.imshow(pred01.squeeze())
    plt.axis('off')
    plt.title('Shiftconv approach')
    plt.show()


demo_shiftconv()


# Derive receptive field from the model
def receptive_field_conv(kernel, stride, n0=1, n_lyrs=1):
    n1 = n0
    for i in range(n_lyrs):
        n1 = kernel + (n1 - 1) * stride
    return n1


def receptive_field_pool(kernel, n0=1, n_lyrs=1):
    n1 = n0 * kernel ** n_lyrs
    return n1


def receptive_field_pool(kernel, n0=1, shift_n=0, n_lyrs=1):
    n1 = n0 * kernel ** n_lyrs
    if shift_n > 1:  # shift_n is the Nst pooling lyr
        s = 1 + 2 ** shift_n
    else:
        s = shift_n
    return n1, s


def receptive_field_up(pl_size, n0=1):
    if n0 == 1:
        return n0
    else:
        return numpy.ceil(n0 * 1 / pl_size).astype(int)


def receptive_field_model(model, verbose=False):
    n1 = 1  # starting field size to calculate receptive field
    shift_n = 0  # index of pooling lyr to calc. shift in shiftconv
    s = 0  # shift due to shiftconv
    layers = numpy.copy(model.layers)
    layers = list(layers)
    layers.reverse()
    for layer in layers:
        lyr_type = layer.__class__.__name__
        if 'Padding' in lyr_type:
            shift = True
        if 'Conv' in lyr_type:
            n1 = receptive_field_conv(layer.kernel_size[0], layer.strides[0], n1)
        elif 'Pooling' in lyr_type:
            if shift:
                shift_n += 1
            n1, s = receptive_field_pool(layer.pool_size[0], n1, shift_n)
        elif 'UpSampling' in lyr_type:
            n1 = receptive_field_up(layer.size[0], n1)
        if verbose:
            print(f'{lyr_type}: RF {n1}, shift {s}')
    if shift:
        n1 = (n1 + s) * 2
        print(
            f'Synthetic receptive field is {n1}^2 (with {int(n1/2 + s)}^2 empty space at 4 corners).'
        )
    return n1



# from tensorflow.python.client import device_lib
# import tensorflow as tf
from aydin.util.tf.device import available_device_memory

if __name__ == '__main__':
    # print(device_lib.list_local_devices())
    # print(tf.test.gpu_device_name())
    print(available_device_memory())



import napari

from aydin.io import io
from aydin.io.datasets import (
    download_from_gdrive,
    examples_single,
    download_all_examples,
    datasets_folder,
)


def demo_examples_single():
    for dataset in examples_single:
        print(dataset)


def demo_download():
    print(
        download_from_gdrive(*examples_single.generic_mandrill.value, datasets_folder)
    )


def demo_all_download():
    download_all_examples()


def demo_examples():
    """
    ....
    """

    for example in examples_single:
        example_file_path = example.get_path()

        # print(f"Trying to open and make sense of file {example_file_path}")

        array, metadata = io.imread(example_file_path)

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(array, name='image')


demo_examples()



# flake8: noqa
import time

import napari
import numpy
import numpy as np
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise, add_noise
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.nn import NNRegressor


def demo():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """

    image = normalise(camera().astype(np.float32, copy=False))
    noisy = add_noise(image)

    generator = StandardFeatureGenerator()  # max_level=4)
    regressor = NNRegressor(max_epochs=1)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    start = time.time()
    it.train(noisy)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    # in case of batching we have to do this:
    start = time.time()
    denoised = it.translate(noisy)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    print("noisy", psnr(noisy, image), ssim(noisy, image))
    print("denoised", psnr(denoised, image), ssim(denoised, image))
    # print("denoised_predict", psnr(denoised_predict, image), ssim(denoised_predict, image))

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')
        viewer.add_image(denoised, name='denoised')


demo()



import importlib
import os
import platform
import shutil
import numpy

from aydin.it import classic_denoisers
from aydin.it.base import ImageTranslatorBase
from aydin.it.classic import ImageDenoiserClassic
from aydin.it.transforms.padding import PaddingTransform
from aydin.it.transforms.range import RangeTransform
from aydin.it.transforms.variance_stabilisation import VarianceStabilisationTransform
from aydin.restoration.denoise.base import DenoiseRestorationBase
from aydin.util.log.log import lsection


if os.getenv("BUNDLED_AYDIN") == "1":
    from aydin.it.classic_denoisers.bilateral import (  # noqa: F401
        calibrate_denoise_bilateral,
        denoise_bilateral,
    )
    from aydin.it.classic_denoisers.bmnd import (  # noqa: F401
        calibrate_denoise_bmnd,
        denoise_bmnd,
    )
    from aydin.it.classic_denoisers.butterworth import (  # noqa: F401
        calibrate_denoise_butterworth,
        denoise_butterworth,
    )
    from aydin.it.classic_denoisers.dictionary_fixed import (  # noqa: F401
        calibrate_denoise_dictionary_fixed,
        denoise_dictionary_fixed,
    )
    from aydin.it.classic_denoisers.dictionary_learned import (  # noqa: F401
        calibrate_denoise_dictionary_learned,
        denoise_dictionary_learned,
    )
    from aydin.it.classic_denoisers.gaussian import (  # noqa: F401
        calibrate_denoise_gaussian,
        denoise_gaussian,
    )
    from aydin.it.classic_denoisers.gm import (  # noqa: F401
        calibrate_denoise_gm,
        denoise_gm,
    )
    from aydin.it.classic_denoisers.harmonic import (  # noqa: F401
        calibrate_denoise_harmonic,
        denoise_harmonic,
    )
    from aydin.it.classic_denoisers.lipschitz import (  # noqa: F401
        calibrate_denoise_lipschitz,
        denoise_lipschitz,
    )
    from aydin.it.classic_denoisers.nlm import (  # noqa: F401
        calibrate_denoise_nlm,
        denoise_nlm,
    )
    from aydin.it.classic_denoisers.pca import (  # noqa: F401
        calibrate_denoise_pca,
        denoise_pca,
    )
    from aydin.it.classic_denoisers.spectral import (  # noqa: F401
        calibrate_denoise_spectral,
        denoise_spectral,
    )
    from aydin.it.classic_denoisers.tv import (  # noqa: F401
        calibrate_denoise_tv,
        denoise_tv,
    )
    from aydin.it.classic_denoisers.wavelet import (  # noqa: F401
        calibrate_denoise_wavelet,
        denoise_wavelet,
    )


class Classic(DenoiseRestorationBase):
    """Classic Image Denoising.

    Parameters
    ----------
    variant : str
        Opacity of the layer visual, between 0.0 and 1.0.
    use_model : bool
        Flag to choose to train a new model or infer from a
        previously trained model. By default it is None.
    input_model_path : string
        Path to model that is desired to be used for inference.
        By default it is None.
    """

    disabled_modules = ["bilateral", "bmnd"]

    def __init__(
        self,
        *,
        variant: str = 'butterworth',
        use_model=None,
        input_model_path=None,
        lower_level_args=None,
        it_transforms=None,
    ):
        super().__init__()
        self.lower_level_args = lower_level_args

        self.backend = variant

        self.input_model_path = input_model_path
        self.use_model_flag = use_model
        self.model_folder_path = None

        self.it = None
        self.it_transforms = (
            [
                {"class": RangeTransform, "kwargs": {}},
                {"class": PaddingTransform, "kwargs": {}},
                {"class": VarianceStabilisationTransform, "kwargs": {}},
            ]
            if it_transforms is None
            else it_transforms
        )

        self.has_less_than_one_million_voxels = False
        self.has_less_than_one_trillion_voxels = True
        self.number_of_dims = -1

    @property
    def configurable_arguments(self):
        """Returns the configurable arguments that will be exposed
        on GUI and CLI.
        """
        arguments = {}

        # Methods
        method_modules = self.get_implementations_in_a_module(classic_denoisers)

        # Remove the disabled modules
        method_modules = [
            module
            for module in method_modules
            if module.name not in self.disabled_modules
        ]

        if platform.system() == "Darwin":
            for module in method_modules:
                if module.name in ["spectral", "dictionary_learned"]:
                    method_modules.remove(module)

        for module in method_modules:
            calibration_args = self.get_function_implementation_kwonlyargs(
                classic_denoisers, module, "calibrate_denoise_" + module.name
            )
            for idx, arg_name in enumerate(calibration_args["arguments"]):
                if arg_name == "display_images":
                    calibration_args["arguments"].remove(arg_name)
                    calibration_args["defaults"] = tuple(
                        x
                        for id, x in enumerate(calibration_args["defaults"])
                        if id != idx
                    )
                    del calibration_args["annotations"][arg_name]

            calibration_args["backend"] = module.name
            arguments["Classic-" + module.name] = {"calibration": calibration_args}

        return arguments

    @property
    def implementations(self):
        """Returns the list of discovered implementations for given method."""
        method_modules = self.get_implementations_in_a_module(classic_denoisers)

        # Remove the disabled modules
        method_modules = [
            module
            for module in method_modules
            if module.name not in self.disabled_modules
        ]

        if platform.system() == "Darwin":
            for module in method_modules:
                if module.name in ["spectral", "dictionary_learned"]:
                    method_modules.remove(module)

        return ["Classic-" + x.name for x in method_modules]

    @property
    def implementations_description(self):
        it_classic_description = ImageDenoiserClassic.__doc__.strip()

        descriptions = []

        method_modules = self.get_implementations_in_a_module(classic_denoisers)

        # Remove the disabled modules
        method_modules = [
            module
            for module in method_modules
            if module.name not in self.disabled_modules
        ]

        if platform.system() == "Darwin":
            for module in method_modules:
                if module.name in ["spectral", "dictionary_learned"]:
                    method_modules.remove(module)

        for module in method_modules:
            response = importlib.import_module(
                classic_denoisers.__name__ + '.' + module.name
            )

            elem = response.__getattribute__("denoise_" + module.name)
            descriptions.append(
                it_classic_description
                + ": "
                + module.name
                + "\n\n"
                + elem.__doc__[: elem.__doc__.find("Parameters")].replace(
                    "\n\n", "<br><br>"
                )
            )

        return descriptions

    def stop_running(self):
        """Method to stop running N2S instance"""
        self.it.stop_training()

    def set_image_metrics(self, image_shape):
        """Sets several image metric parameters used internally.

        Parameters
        ----------
        image_shape : tuple

        """
        self.number_of_dims = len(image_shape)
        number_of_voxels = numpy.prod(numpy.array(image_shape))
        self.has_less_than_one_million_voxels = number_of_voxels < 1000000
        self.has_less_than_one_trillion_voxels = number_of_voxels < 1000000000000

    def get_translator(self):
        """Returns the corresponding translator instance for given selections.

        Parameters
        ----------
        feature_generator : FeatureGeneratorBase
        regressor : RegressorBase

        Returns
        -------
        it : ImageTranslatorBase

        """
        # Use a pre-saved model or train a new one from scratch and save it
        if self.use_model_flag:
            # Unarchive the model file and load its ImageTranslator object into self.it
            shutil.unpack_archive(
                self.input_model_path, os.path.dirname(self.input_model_path), "zip"
            )
            it = ImageTranslatorBase.load(self.input_model_path[:-4])
        else:
            if self.lower_level_args is not None:
                method = (
                    self.backend
                    if self.lower_level_args["variant"] is None
                    else self.lower_level_args["variant"].split("-")[1]
                )
                it = ImageDenoiserClassic(
                    method=method,
                    calibration_kwargs=self.lower_level_args["calibration"]["kwargs"],
                )
            else:
                it = ImageDenoiserClassic(method=self.backend)

        return it

    def add_transforms(self):
        if self.it_transforms is not None:
            for transform in self.it_transforms:
                transform_class = transform["class"]
                transform_kwargs = transform["kwargs"]
                self.it.add_transform(transform_class(**transform_kwargs))

    def train(
        self, noisy_image, *, batch_axes=None, chan_axes=None, image_path=None, **kwargs
    ):
        """Method to run training for Noise2Self FGR.

        Parameters
        ----------
        noisy_image : numpy.ndarray
        batch_axes : array_like, optional
            Indices of batch axes.
        chan_axes : array_like, optional
            Indices of channel axes.
        image_path : str

        Returns
        -------
        response : numpy.ndarray

        """
        with lsection("Noise2Self train is starting..."):
            self.set_image_metrics(noisy_image.shape)

            self.it = self.get_translator()

            self.add_transforms()

            # Train a new model
            self.it.train(
                noisy_image,
                noisy_image,
                batch_axes=batch_axes,
                channel_axes=chan_axes,
                train_valid_ratio=kwargs['train_valid_ratio']
                if 'train_valid_ratio' in kwargs
                else 0.1,
                callback_period=kwargs['callback_period']
                if 'callback_period' in kwargs
                else 3,
                jinv=kwargs['jinv'] if 'jinv' in kwargs else None,
            )

            # Save the trained model
            # self.save_model(image_path)  # TODO:  fix the problems here

    def denoise(self, noisy_image, *, batch_axes=None, chan_axes=None, **kwargs):
        """Method to denoise an image with trained Noise2Self FGR.

        Parameters
        ----------
        batch_axes : array_like, optional
            Indices of batch axes.
        chan_axes : array_like, optional
            Indices of channel axes.
        noisy_image : numpy.ndarray

        Returns
        -------
        response : numpy.ndarray

        """
        with lsection("Noise2Self denoise is starting..."):

            # Predict the resulting image
            response = self.it.translate(
                noisy_image,
                batch_axes=batch_axes,
                channel_axes=chan_axes,
                tile_size=kwargs['tile_size'] if 'tile_size' in kwargs else None,
            )

            response = response.astype(noisy_image.dtype, copy=False)

            return response


def classic_denoise(noisy, *, batch_axes=None, chan_axes=None, variant=None):
    """Method to denoise an image with Classic denoising restoration module.

    Parameters
    ----------
    noisy : numpy.ndarray
        Image to denoise
    batch_axes : array_like, optional
        Indices of batch axes.
    chan_axes : array_like, optional
        Indices of channel axes.
    variant : str
        Algorithm variant.

    Returns
    -------
    Denoised image : numpy.ndarray

    """
    # Run and save the result
    classic = Classic(variant=variant)

    # Train
    classic.train(noisy, batch_axes=batch_axes, chan_axes=chan_axes)

    # Denoise
    denoised = classic.denoise(noisy, batch_axes=batch_axes, chan_axes=chan_axes)

    return denoised



import numpy
from scipy.ndimage import median_filter
from skimage.exposure import rescale_intensity

from aydin.features.groups.median import MedianFeatures
from aydin.io.datasets import camera


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def test_median_feature_group():
    # get image:
    image = n(camera().astype(numpy.float32))

    # Instantiates median features:
    radii = [1, 2, 4, 8]
    medians = MedianFeatures(radii=radii)
    assert medians.num_features(image.ndim) == 4

    # Check receptive field radius:
    assert medians.receptive_field_radius == 8

    # Set image:
    medians.prepare(image)

    # compute features and check their valididty:
    feature = numpy.empty_like(image)
    for index in range(medians.num_features(image.ndim)):
        medians.compute_feature(index=index, feature=feature)
        radius = radii[index]
        assert (feature == median_filter(image, size=2 * radius + 1)).all()



import importlib
from typing import Optional
import numpy

from aydin.it import classic_denoisers
from aydin.it.base import ImageTranslatorBase
from aydin.util.log.log import lsection, lprint


class ImageDenoiserClassic(ImageTranslatorBase):
    """
    Classic Image Denoiser
    """

    def __init__(
        self,
        method: str = "butterworth",
        main_channel: Optional[int] = None,
        max_voxels_for_training: Optional[int] = None,
        calibration_kwargs=None,
        **kwargs,
    ):
        """Constructs a Classic image denoiser.

        Parameters
        ----------
        method: str
            Name of classical denoising method.

        main_channel: optional int
            By default the denoiser is calibrated per channel.
            To speed up denoising you can pick one channel index
            to use during calibration and used to denoise all channels.

        max_voxels_for_training : int, optional
            Maximum number of the voxels that can be
            used for training.

        kwargs : dict
            Keyword arguments.
        """
        super().__init__(**kwargs)

        self.calibration_kwargs = (
            {} if calibration_kwargs is None else calibration_kwargs
        )

        response = importlib.import_module(classic_denoisers.__name__ + '.' + method)

        self.calibration_function = response.__getattribute__(
            "calibrate_denoise_" + method
        )

        self.max_voxels_for_training = max_voxels_for_training

        self._memory_needed = 0

        self.main_channel = main_channel

        with lsection("Classic image translator"):
            lprint(f"method: {method}")
            lprint(f"main channel: {main_channel}")

    def save(self, path: str):
        """Saves a 'all-batteries-included' image translation model at a given path (folder).

        Parameters
        ----------
        path : str
            path to save to

        Returns
        -------
        frozen

        """
        with lsection(f"Saving 'classic' image denoiser to {path}"):
            frozen = super().save(path)
            # TODO: implement!

        return frozen

    def _load_internals(self, path: str):
        with lsection(f"Loading 'classic' image denoiser from {path}"):
            # TODO: implement!
            pass
            # self.best_parameters = FeatureGeneratorBase.load(path)

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        # TODO: implement!
        return state

    def _train(
        self, input_image, target_image, train_valid_ratio, callback_period, jinv
    ):
        with lsection(
            f"Training image translator from image of shape {input_image.shape}:"
        ):
            shape = input_image.shape
            num_channels = shape[1]

            self.best_parameters = []
            self.denoising_functions = []

            # We calibrate per channel
            for channel_index in range(num_channels):
                lprint(f'Calibrating denoiser on channel {channel_index}')
                channel_image = input_image[:, channel_index]

                # for a given channel we find the best batch to use:
                # We pick the batch with highest variance:
                variance_list = [numpy.std(i) for i in channel_image]
                batch_index = variance_list.index(max(variance_list))

                # We pick that batch image:
                image = channel_image[batch_index]

                (
                    denoising_function,
                    best_parameters,
                    memory_requirements,
                ) = self.calibration_function(image, **self.calibration_kwargs)

                # Add obtained best parameters to the list per channel:
                self.denoising_functions.append(denoising_function)
                self.best_parameters.append(best_parameters)
                self._memory_needed = memory_requirements

    def _estimate_memory_needed_and_available(self, image):
        """

        Parameters
        ----------
        image

        Returns
        -------

        """
        _, available = super()._estimate_memory_needed_and_available(image)

        return self._memory_needed, available

    def _translate(self, input_image, image_slice=None, whole_image_shape=None):
        """Internal method that translates an input image on the basis of the trained model.

        :param input_image: input image
        :param batch_dims: batch dimensions
        :return:
        """
        shape = input_image.shape
        num_batches = shape[0]
        num_channels = shape[1]

        denoised_image = numpy.empty_like(input_image)

        for batch_index in range(num_batches):
            for channel_index in range(num_channels):
                lprint(
                    f'Denoising image for batch: {batch_index} and channel: {channel_index}'
                )
                best_parameters = self.best_parameters[channel_index]
                denoising_function = self.denoising_functions[channel_index]
                image = input_image[batch_index, channel_index]
                denoised_image[batch_index, channel_index] = denoising_function(
                    image, **best_parameters
                )

        return denoised_image



import pytest
from os import path
from os.path import join

import numpy

from aydin.io import io
from aydin.io.datasets import examples_single
from aydin.io.folders import get_temp_folder
from aydin.io.io import imread, mapped_tiff, imwrite


def test_analysis():

    for example in examples_single:

        example_file_path = example.get_path()

        _, analysis_result = io.imread(example_file_path)

        print(analysis_result)


@pytest.mark.heavy
def test_opening_examples():

    for example in examples_single:

        example_file_path = example.get_path()

        # print(f"Trying to open and make sense of file {example_file_path}")

        array, metadata = io.imread(example_file_path)

        if array is not None:
            print(
                f"dataset: {example.value[1]}, shape:{array.shape}, dtype:{array.dtype} "
            )
        else:
            print(f"Cannot open dataset: {example.value[1]}")


def test_imwrite():
    input_path = examples_single.janelia_flybrain.get_path()
    assert path.exists(input_path)

    array, metadata = imread(input_path)

    temp_file = join(get_temp_folder(), "test_imwrite.tif")

    print("Writting file...")
    imwrite(array, temp_file)
    print("Done...")

    print("Reading back...")
    array2, _ = imread(temp_file)

    assert numpy.all(array == array2)

    input_path = examples_single.generic_lizard.get_path()
    assert path.exists(input_path)

    array, metadata = imread(input_path)
    temp_file = join(get_temp_folder(), "test_imwrite.png")

    print("Writting file...")
    imwrite(array, temp_file)
    print("Done...")

    print("Reading back...")
    array2, _ = imread(temp_file)

    assert numpy.all(array == array2)


def test_mapped_tiff():

    input_path = examples_single.janelia_flybrain.get_path()

    assert path.exists(input_path)

    array, metadata = imread(input_path)

    temp_file = join(get_temp_folder(), "test_imwrite.tif")

    # We use a generator that takes care of everything:
    with mapped_tiff(temp_file, shape=array.shape, dtype=array.dtype) as tiff_array:

        # We write here the file
        print("Writting file...")
        tiff_array[...] = array

    print("Done...")

    print("Reading back...")
    array2, _ = imread(temp_file)

    assert numpy.all(array == array2)



# flake8: noqa
import time

import numpy
import numpy as np
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim
from skimage.util import random_noise

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import normalise
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.lgbm import LGBMRegressor
from aydin.util.log.log import Log


def demo():
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True

    image = camera().astype(np.float32)  # newyork()[256:-256, 256:-256]
    image = normalise(image)

    intensity = 5
    np.random.seed(0)
    noisy = np.random.poisson(image * intensity) / intensity
    noisy = random_noise(noisy, mode='gaussian', var=0.01, seed=0)
    noisy = noisy.astype(np.float32)

    generator = StandardFeatureGenerator(max_level=10)
    regressor = LGBMRegressor()

    it = ImageTranslatorFGR(
        feature_generator=generator, regressor=regressor, max_memory_usage_ratio=0.0001
    )

    start = time.time()
    it.train(noisy, noisy)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    # in case of batching we have to do this:
    start = time.time()
    denoised = it.translate(noisy)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)

    print("noisy       :", psnr(image, noisy), ssim(noisy, image))
    print("denoised    :", psnr(image, denoised), ssim(denoised, image))

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(normalise(image), name='image')
        viewer.add_image(normalise(noisy), name='noisy')
        viewer.add_image(normalise(denoised), name='denoised')


demo()



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QStackedWidget,
    QListWidget,
)

from aydin.gui._qt.custom_widgets.denoise_tab_method import DenoiseTabMethodWidget
from aydin.gui._qt.custom_widgets.horizontal_line_break_widget import (
    QHorizontalLineBreakWidget,
)
from aydin.gui._qt.custom_widgets.readmoreless_label import QReadMoreLessLabel
from aydin.restoration.denoise.util.denoise_utils import (
    get_list_of_denoiser_implementations,
)


class DenoiseTab(QWidget):
    """
    Now it is time to denoise the previously selected and cropped images.
    <br><br>
    Aydin comes with a growing variety of self-supervised, auto-tuned, and unsupervised denoising algorithms,
    each with their own strengths and weaknesses in terms of speed, denoising performance, artifacts, and propensity
    to hallucinate unsubstantiated details. We recommend you check our
    <a href='https://royerlab.github.io/aydin/use_cases/introduction.html'>use cases</a> to learn how to
    choose the best algorithm and parameters for your image(s).

    <moreless>
    Quick guide: The first algorithm to try is 'Butterworth' which is remarkably simple, fast, and sometimes
    embarrassingly effective compared to more sophisticated methods. Next you can try the Gaussian-Median mixed
    denoiser (gm) which is in some cases quite effective. 'Spectral' can give extremely good results too but can take
    longer, as do in general patch and dictionary based methods.
    <split>
    Our own favourite is a novel variant on the Noise2Self theme which relies on carefully crafted features and
    gradient boosting (N2S-FGR-cb or -lgbm). CNN-based Noise2Self denoising is also available but is currently on of
    our least favourites because of its propensity to hallucinate detail, slowness, and overall worse performance. In
    fine, there is no silver bullet, there is not a single denoising algorithm that can tackle all denoising
    challenges, instead you need to choose and play with a variety of algorithms to find the one that will fit you
    needs both in terms of processing speed, visual appearance, and downstream analysis constraints. Denoising is a
    form of image analysis that consists in separating signal from noise, with the definition of signal and noise
    being to some extent, subjective and context dependent.
    """

    def __init__(self, parent):
        super(DenoiseTab, self).__init__(parent)
        self.parent = parent

        self.tab_layout = QVBoxLayout()
        self.tab_layout.setAlignment(Qt.AlignTop)

        self.tab_layout.addWidget(QReadMoreLessLabel(self, self.__doc__))

        # Horizontal Line Break
        self.tab_layout.addWidget(QHorizontalLineBreakWidget(self))

        self.leftlist = QListWidget()

        (
            backend_options,
            backend_options_descriptions,
        ) = get_list_of_denoiser_implementations()

        backend_options, backend_options_descriptions = (
            list(t)
            for t in zip(*sorted(zip(backend_options, backend_options_descriptions)))
        )

        self.stacked_widget = QStackedWidget(self)

        default_option_index = 0
        for idx, (backend_option, description) in enumerate(
            zip(backend_options, backend_options_descriptions)
        ):
            self.leftlist.insertItem(idx, backend_option)
            # self.leftlist.item(idx).setToolTip(tooltip)

            self.stacked_widget.addWidget(
                DenoiseTabMethodWidget(
                    self, name=backend_option, description=description
                )
            )

            if backend_option == "Classic-butterworth":
                default_option_index = idx

        self.leftlist.item(default_option_index).setSelected(True)
        self.change_current_method(default_option_index)
        self.leftlist.currentRowChanged.connect(self.change_current_method)

        hbox = QHBoxLayout()
        hbox.addWidget(self.leftlist, 15)
        hbox.addWidget(self.stacked_widget, 85)

        self.tab_layout.addLayout(hbox)

        self.setLayout(self.tab_layout)

    def change_current_method(self, new_index):
        self.stacked_widget.setCurrentIndex(new_index)

    @property
    def selected_backend(self):
        return self.stacked_widget.currentWidget().name

    @property
    def current_backend_qwidget(self):
        return self.stacked_widget.currentWidget()

    @property
    def lower_level_args(self):
        return self.stacked_widget.currentWidget().lower_level_args()
